@@ -1,7 +1,7 @@
 /*
  * dict.c	Routines to read the dictionary file.
  *
- * Version:	$Id: dict.c,v 1.39.2.1 2003/09/16 18:34:36 phampson Exp $
+ * Version:	$Id: dict.c,v 1.50.2.3 2004/09/10 15:06:44 aland Exp $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -20,7 +20,7 @@
  * Copyright 2000  The FreeRADIUS server project
  */
 
-static const char rcsid[] = "$Id: dict.c,v 1.39.2.1 2003/09/16 18:34:36 phampson Exp $";
+static const char rcsid[] = "$Id: dict.c,v 1.50.2.3 2004/09/10 15:06:44 aland Exp $";
 
 #include	"autoconf.h"
 
@@ -28,17 +28,37 @@ static const char rcsid[] = "$Id: dict.c,v 1.39.2.1 2003/09/16 18:34:36 phampson
 #include	<ctype.h>
 #include	<string.h>
 
-#if HAVE_MALLOC_H
+#ifdef HAVE_MALLOC_H
 #include	<malloc.h>
 #endif
 
 #include	"libradius.h"
 #include	"missing.h"
 
-static DICT_ATTR	*dictionary_attributes = NULL;
-static DICT_VALUE	*dictionary_values = NULL;
+/*
+ *	There are very few vendors, and they're looked up only when we
+ *	read the dictionaries.  So it's OK to have a singly linked
+ *	list here.
+ */
 static DICT_VENDOR	*dictionary_vendors = NULL;
 
+static rbtree_t *attributes_byname = NULL;
+static rbtree_t *attributes_byvalue = NULL;
+
+static rbtree_t *values_byvalue = NULL;
+static rbtree_t *values_byname = NULL;
+
+typedef struct value_fixup_t {
+	char		attrstr[40];
+	DICT_VALUE	*dval;
+	struct value_fixup_t *next;
+} value_fixup_t;
+
+/*
+ *	So VALUEs in the dictionary can have forward references.
+ */
+static value_fixup_t *value_fixup = NULL;
+
 static const LRAD_NAME_NUMBER type_table[] = {
 	{ "string",	PW_TYPE_STRING },
 	{ "integer",	PW_TYPE_INTEGER },
@@ -46,6 +66,9 @@ static const LRAD_NAME_NUMBER type_table[] = {
 	{ "date",	PW_TYPE_DATE },
 	{ "abinary",	PW_TYPE_ABINARY },
 	{ "octets",	PW_TYPE_OCTETS },
+	{ "ifid",	PW_TYPE_IFID },
+	{ "ipv6addr",	PW_TYPE_IPV6ADDR },
+	{ "ipv6prefix", PW_TYPE_IPV6PREFIX },
 	{ NULL, 0 }
 };
 
@@ -55,37 +78,38 @@ static const LRAD_NAME_NUMBER type_table[] = {
  *	These attributes are referenced a LOT, especially during
  *	decoding of the on-the-wire packets.  It's useful to keep a
  *	cache of their dictionary entries, so looking them up is
- *	O(1), instead of O(N).  (N==number of dictionary entries...)
+ *	O(1), instead of O(log(N)).  (N==number of dictionary entries...)
  */
 static DICT_ATTR *base_attributes[256];
 
 /*
  *	Free the dictionary_attributes and dictionary_values lists.
  */
-static void dict_free(void)
+void dict_free(void)
 {
-	DICT_ATTR	*dattr, *anext;
-	DICT_VALUE	*dval, *vnext;
 	DICT_VENDOR	*dvend, *enext;
 
-	for (dattr = dictionary_attributes; dattr; dattr = anext) {
-		anext = dattr->next;
-		free(dattr);
-	}
-	for (dval = dictionary_values; dval; dval = vnext) {
-		vnext = dval->next;
-		free(dval);
-	}
+	memset(base_attributes, 0, sizeof(base_attributes));
+
 	for (dvend = dictionary_vendors; dvend; dvend = enext) {
 		enext = dvend->next;
 		free(dvend);
 	}
 
-	dictionary_attributes = NULL;
-	dictionary_values = NULL;
 	dictionary_vendors = NULL;
 
-	memset(base_attributes, 0, sizeof(base_attributes));
+	/*
+	 *	Free the tree of attributes by name and value.
+	 */
+	rbtree_free(attributes_byname);
+	rbtree_free(attributes_byvalue);
+	attributes_byname = NULL;
+	attributes_byvalue = NULL;
+
+	rbtree_free(values_byname);
+	rbtree_free(values_byvalue);
+	values_byname = NULL;
+	values_byvalue = NULL;
 }
 
 /*
@@ -100,7 +124,7 @@ int dict_addvendor(const char *name, int value)
 		return -1;
 	}
 
-	if (strlen(name) > (sizeof(vval->vendorname) -1)) {
+	if (strlen(name) > (sizeof(vval->name) -1)) {
 		librad_log("dict_addvendor: vendor name too long");
 		return -1;
 	}
@@ -109,7 +133,7 @@ int dict_addvendor(const char *name, int value)
 		librad_log("dict_addvendor: out of memory");
 		return -1;
 	}
-	strcpy(vval->vendorname, name);
+	strcpy(vval->name, name);
 	vval->vendorpec  = value;
 
 	/* Insert at front. */
@@ -119,7 +143,11 @@ int dict_addvendor(const char *name, int value)
 	return 0;
 }
 
-int dict_addattr(const char *name, int vendor, int type, int value, ATTR_FLAGS flags)
+/*
+ *	Add an attribute to the dictionary.
+ */
+int dict_addattr(const char *name, int vendor, int type, int value,
+		 ATTR_FLAGS flags)
 {
 	static int      max_attr = 0;
 	DICT_ATTR	*attr;
@@ -152,6 +180,10 @@ int dict_addattr(const char *name, int vendor, int type, int value, ATTR_FLAGS f
 		}
 	}
 
+	if (value >= 65536) {
+		librad_log("dict_addattr: ATTRIBUTE has invalid number.");
+		return -1;
+	}
 
 	/*
 	 *	Create a new attribute for the list
@@ -165,9 +197,35 @@ int dict_addattr(const char *name, int vendor, int type, int value, ATTR_FLAGS f
 	attr->type = type;
 	attr->flags = flags;
 
-	if (vendor) {
-		attr->attr |= (vendor << 16);
-	} else if ((attr->attr >= 0) && (attr->attr < 256)) {
+	if (vendor) attr->attr |= (vendor << 16);
+
+	/*
+	 *	Insert the attribute, only if it's not a duplicate.
+	 */
+	if (rbtree_insert(attributes_byname, attr) == 0) {
+		DICT_ATTR	*a;
+
+		/*
+		 *	If the attribute has identical number, then
+		 *	ignore the duplicate.
+		 */
+		a = rbtree_finddata(attributes_byname, attr);
+		if (a && (strcasecmp(a->name, attr->name) == 0)) {
+			if (a->attr != attr->attr) {
+				librad_log("dict_addattr: Duplicate attribute name %s", name);
+				return -1;
+			}
+
+			/*
+			 *	Same name, same attr, maybe the
+			 *	flags and/or type is different.
+			 *	Let the new value over-ride the
+			 *	old one.
+			 */
+		}
+	}
+
+	if ((attr->attr >= 0) && (attr->attr < 256)) {
 		/*
 		 *	If it's an on-the-wire base attribute,
 		 *	then keep a quick reference to it, for speed.
@@ -176,16 +234,27 @@ int dict_addattr(const char *name, int vendor, int type, int value, ATTR_FLAGS f
 	}
 
 	/*
-	 *	Add to the front of the list, so that
-	 *	values at the end of the file override
-	 *	those in the begin.
+	 *	Insert the SAME pointer (not free'd when this tree is
+	 *	deleted), into another tree.
+	 *
+	 *	If the newly inserted entry is a duplicate of an existing
+	 *	entry, then the old entry is tossed, and the new one
+	 *	replaces it.  This behaviour is configured in the
+	 *	rbtree_create() function.
+	 *
+	 *	We want this behaviour because we want OLD names for
+	 *	the attributes to be read from the configuration
+	 *	files, but when we're printing them, (and looking up
+	 *	by value) we want to use the NEW name.
 	 */
-	attr->next = dictionary_attributes;
-	dictionary_attributes = attr;
-	
+	rbtree_insert(attributes_byvalue, attr);
+
 	return 0;
 }
 
+/*
+ *	Add a value for an attribute to the dictionary.
+ */
 int dict_addvalue(const char *namestr, char *attrstr, int value)
 {
 	DICT_ATTR	*dattr;
@@ -196,18 +265,14 @@ int dict_addvalue(const char *namestr, char *attrstr, int value)
 		return -1;
 	}
 
-	if (strlen(attrstr) > (sizeof(dval->attrname) -1)) {
-		librad_log("dict_addvalue: attribute name too long");
-		return -1;
-	}
-
 	if ((dval = (DICT_VALUE *)malloc(sizeof(DICT_VALUE))) == NULL) {
 		librad_log("dict_addvalue: out of memory");
 		return -1;
 	}
+	memset(dval, 0, sizeof(*dval));
 
 	strcpy(dval->name, namestr);
-	strcpy(dval->attrname, attrstr);
+	dval->value = value;
 
 	/*
 	 *	Remember which attribute is associated with this
@@ -217,13 +282,50 @@ int dict_addvalue(const char *namestr, char *attrstr, int value)
 	if (dattr) {
 		dval->attr = dattr->attr;
 	} else {
-		dval->attr = 0;
+		value_fixup_t *fixup;
+		
+		fixup = (value_fixup_t *) malloc(sizeof(*fixup));
+		if (!fixup) {
+			librad_log("dict_addvalue: out of memory");
+			return -1;
+		}
+		memset(fixup, 0, sizeof(*fixup));
+
+		strNcpy(fixup->attrstr, attrstr, sizeof(fixup->attrstr));
+		fixup->dval = dval;
+		
+		/*
+		 *	Insert to the head of the list.
+		 */
+		fixup->next = value_fixup;
+		value_fixup = fixup;
+
+		return 0;
 	}
-	dval->value = value;
-	
-	/* Insert at front. */
-	dval->next = dictionary_values;
-	dictionary_values = dval;
+
+	/*
+	 *	Add the value into the dictionary.
+	 */
+	if (rbtree_insert(values_byname, dval) == 0) {
+		if (dattr) {
+			DICT_VALUE *dup;
+			
+			/*
+			 *	Suppress duplicates with the same
+			 *	name and value.  There are lots in
+			 *	dictionary.ascend.
+			 */
+			dup = dict_valbyname(dattr->attr, namestr);
+			if (dup && (dup->value == dval->value)) {
+				free(dval);
+				return 0;
+			}
+		}
+
+		librad_log("dict_addvalue: Duplicate value name %s for attribute %s", namestr, attrstr);
+		return -1;
+	}
+	rbtree_insert(values_byvalue, dval);
 
 	return 0;
 }
@@ -231,7 +333,9 @@ int dict_addvalue(const char *namestr, char *attrstr, int value)
 /*
  *	Process the ATTRIBUTE command
  */
-static int process_attribute(const char* fn, const int line, const int block_vendor, const char* data) {
+static int process_attribute(const char* fn, const int line,
+			     const int block_vendor, const char* data)
+{
 	int		vendor;
 	char		namestr[256];
 	char		valstr[256];
@@ -292,7 +396,7 @@ static int process_attribute(const char* fn, const int line, const int block_ven
 			 flags.has_tag = 1;
 		}
 		else if (strncmp(s, "len+=", 5) == 0 ||
-			 strncmp(s, "len-=", 5) == 0) { 
+			 strncmp(s, "len-=", 5) == 0) {
 			  /* Length difference, to accomodate
 			     braindead NASes & their vendors */
 			  flags.len_disp = strtol(s + 5, &c, 0);
@@ -309,7 +413,7 @@ static int process_attribute(const char* fn, const int line, const int block_ven
 			  /* Encryption method, defaults to 0 (none).
 			     Currently valid is just type 2,
 			     Tunnel-Password style, which can only
-			     be applied to strings. */    
+			     be applied to strings. */
 			  flags.encrypt = strtol(s + 8, &c, 0);
 			  if (*c) {
 				librad_log( "dict_init: %s[%d] invalid option %s",
@@ -321,10 +425,10 @@ static int process_attribute(const char* fn, const int line, const int block_ven
 			  /* Must be a vendor 'flag'... */
 			  if (strncmp(s, "vendor=", 5) == 0) {
 				/* New format */
-				s += 5;   
+				s += 5;
 			  }
-	 
-			  vendor = dict_vendorname(s);
+
+			  vendor = dict_vendorbyname(s);
 			  if (!vendor) {
 				librad_log( "dict_init: %s[%d]: unknown vendor %s",
 					   fn, line, optstr);
@@ -355,7 +459,8 @@ static int process_attribute(const char* fn, const int line, const int block_ven
 /*
  *	Process the VALUE command
  */
-static int process_value(const char* fn, const int line, const int block_vendor, const char* data) {
+static int process_value(const char* fn, const int line, const char* data)
+{
 	char	namestr[256];
 	char	valstr[256];
 	char	attrstr[256];
@@ -386,7 +491,7 @@ static int process_value(const char* fn, const int line, const int block_vendor,
 		sscanf(valstr, "%i", &value);
 
 	if (dict_addvalue(namestr, attrstr, value) < 0) {
-		librad_log("dict_init: %s[%d]: %s", 
+		librad_log("dict_init: %s[%d]: %s",
 			   fn, line, librad_errstr);
 		return -1;
 	}
@@ -398,7 +503,8 @@ static int process_value(const char* fn, const int line, const int block_vendor,
 /*
  *	Process the VENDOR command
  */
-static int process_vendor(const char* fn, const int line, const int block_vendor, const char* data) {
+static int process_vendor(const char* fn, const int line, const char* data)
+{
 	char	valstr[256];
 	char	attrstr[256];
 	int	value;
@@ -434,7 +540,8 @@ static int process_vendor(const char* fn, const int line, const int block_vendor
 /*
  *	Initialize the dictionary.
  */
-static int my_dict_init(const char *dir, const char *fn, const char *src_file, int src_line)
+static int my_dict_init(const char *dir, const char *fn,
+			const char *src_file, int src_line)
 {
 	FILE	*fp;
 	char 	dirtmp[256];
@@ -532,19 +639,18 @@ static int my_dict_init(const char *dir, const char *fn, const char *src_file, i
 		 *	Process VALUE lines.
 		 */
 		if (strcasecmp(keyword, "VALUE") == 0) {
-			if (process_value(fn, line, block_vendor, data) == -1) {
+			if (process_value(fn, line, data) == -1) {
 				fclose(fp);
 				return -1;
 			}
 			continue;
 		}
 
-
 		/*
 		 *	Process VENDOR lines.
 		 */
 		if (strcasecmp(keyword, "VENDOR") == 0) {
-			if (process_vendor(fn, line, block_vendor, data) == -1) {
+			if (process_vendor(fn, line, data) == -1) {
 				fclose(fp);
 				return -1;
 			}
@@ -561,7 +667,7 @@ static int my_dict_init(const char *dir, const char *fn, const char *src_file, i
 				return -1;
 			}
 
-			vendor = dict_vendorname(optstr);
+			vendor = dict_vendorbyname(optstr);
 			if (!vendor) {
 				librad_log(
 					"dict_init: %s[%d]: unknown vendor %s",
@@ -583,7 +689,7 @@ static int my_dict_init(const char *dir, const char *fn, const char *src_file, i
 				return -1;
 			}
 
-			vendor = dict_vendorname(optstr);
+			vendor = dict_vendorbyname(optstr);
 			if (!vendor) {
 				librad_log(
 					"dict_init: %s[%d]: unknown vendor %s",
@@ -616,29 +722,139 @@ static int my_dict_init(const char *dir, const char *fn, const char *src_file, i
 	return 0;
 }
 
+/*
+ *	Callbacks for red-black trees.
+ */
+static int attrname_cmp(const void *a, const void *b)
+{
+	return strcasecmp(((const DICT_ATTR *)a)->name,
+			  ((const DICT_ATTR *)b)->name);
+}
+
+/*
+ *	Return: < 0  if a < b,
+ *	        == 0 if a == b
+ */
+static int attrvalue_cmp(const void *a, const void *b)
+{
+	return (((const DICT_ATTR *)a)->attr -
+		((const DICT_ATTR *)b)->attr);
+}
+
+/*
+ *	Compare values by name, keying off of the attribute number,
+ *	and then the value name.
+ */
+static int valuename_cmp(const void *a, const void *b)
+{
+	int rcode;
+	rcode = (((const DICT_VALUE *)a)->attr -
+		 ((const DICT_VALUE *)b)->attr);
+	if (rcode != 0) return rcode;
+
+return strcasecmp(((const DICT_VALUE *)a)->name,
+			  ((const DICT_VALUE *)b)->name);
+}
+
+/*
+ *	Compare values by value, keying off of the attribute number,
+ *	and then the value number.
+ */
+static int valuevalue_cmp(const void *a, const void *b)
+{
+	int rcode;
+	rcode = (((const DICT_VALUE *)a)->attr -
+		 ((const DICT_VALUE *)b)->attr);
+	if (rcode != 0) return rcode;
+
+	return (((const DICT_VALUE *)a)->value -
+		 ((const DICT_VALUE *)b)->value);
+}
+
 /*
  *	Initialize the directory, then fix the attr member of
  *	all attributes.
  */
 int dict_init(const char *dir, const char *fn)
 {
-	DICT_ATTR	*attr;
-	DICT_VALUE	*dval;
-
 	dict_free();
 
+	/*
+	 *	Create the tree of attributes by name.   There MAY NOT
+	 *	be multiple attributes of the same name.
+	 *
+	 *	Each attribute is malloc'd, so the free function is free.
+	 */
+	attributes_byname = rbtree_create(attrname_cmp, free, 0);
+	if (!attributes_byname) {
+		return -1;
+	}
+
+	/*
+	 *	Create the tree of attributes by value.  There MAY
+	 *	be attributes of the same value.  If there are, we
+	 *	pick the latest one.
+	 */
+	attributes_byvalue = rbtree_create(attrvalue_cmp, NULL, 1);
+	if (!attributes_byvalue) {
+		return -1;
+	}
+
+	values_byname = rbtree_create(valuename_cmp, free, 0);
+	if (!values_byname) {
+		return -1;
+	}
+
+	values_byvalue = rbtree_create(valuevalue_cmp, NULL, 1);
+	if (!values_byvalue) {
+		return -1;
+	}
+
+	value_fixup = NULL;	/* just to be safe. */
+
 	if (my_dict_init(dir, fn, NULL, 0) < 0)
 		return -1;
 
-	for (dval = dictionary_values; dval; dval = dval->next) {
-		if (dval->attr != 0)
-			continue;
-		if ((attr = dict_attrbyname(dval->attrname)) == NULL) {
-		librad_log("dict_init: VALUE %s for unknown ATTRIBUTE %s",
-			dval->name, dval->attrname);
-			return -1;
+	if (value_fixup) {
+		DICT_ATTR *a;
+		value_fixup_t *this, *next;
+
+		for (this = value_fixup; this != NULL; this = next) {
+			next = this->next;
+
+			a = dict_attrbyname(this->attrstr);
+			if (!a) {
+				librad_log(
+					"dict_init: No ATTRIBUTE \"%s\" defined for VALUE \"%s\"",
+					this->attrstr, this->dval->name);
+				return -1; /* leak, but they should die... */
+			}
+
+			this->dval->attr = a->attr;
+
+			/*
+			 *	Add the value into the dictionary.
+			 */
+			if (rbtree_insert(values_byname, this->dval) == 0) {
+				librad_log("dict_addvalue: Duplicate value name %s for attribute %s", this->dval->name, a->name);
+				return -1;
+			}
+			
+			/*
+			 *	Allow them to use the old name, but
+			 *	prefer the new name when printing
+			 *	values.
+			 */
+			if (!rbtree_find(values_byvalue, this->dval)) {
+				rbtree_insert(values_byvalue, this->dval);
+			}
+			free(this);
+
+			/*
+			 *	Just so we don't lose track of things.
+			 */
+			value_fixup = next;
 		}
-		dval->attr = attr->attr;
 	}
 
 	return 0;
@@ -647,54 +863,48 @@ int dict_init(const char *dir, const char *fn)
 /*
  *	Get an attribute by its numerical value.
  */
-DICT_ATTR * dict_attrbyvalue(int val)
+DICT_ATTR *dict_attrbyvalue(int val)
 {
-	DICT_ATTR	*a;
-
 	/*
 	 *	If it's an on-the-wire base attribute, return
 	 *	the cached value for it.
 	 */
 	if ((val >= 0) && (val < 256)) {
 		return base_attributes[val];
+
+	} else {
+		DICT_ATTR myattr;
+
+		myattr.attr = val;
+		return rbtree_finddata(attributes_byvalue, &myattr);
 	}
 
-	for (a = dictionary_attributes; a; a = a->next) {
-		if (a->attr == val)
-			return a;
-	}
-
-	return NULL;
+	return NULL;		/* never reached, but useful */
 }
 
 /*
  *	Get an attribute by its name.
  */
-DICT_ATTR * dict_attrbyname(const char *name)
+DICT_ATTR *dict_attrbyname(const char *name)
 {
-	DICT_ATTR	*a;
+	DICT_ATTR myattr;
 
-	for (a = dictionary_attributes; a; a = a->next) {
-		if (strcasecmp(a->name, name) == 0)
-			return a;
-	}
+	strNcpy(myattr.name, name, sizeof(myattr.name));
 
-	return NULL;
+	return rbtree_finddata(attributes_byname, &myattr);
 }
 
 /*
  *	Associate a value with an attribute and return it.
  */
-DICT_VALUE * dict_valbyattr(int attr, int val)
+DICT_VALUE *dict_valbyattr(int attr, int val)
 {
-	DICT_VALUE	*v;
+	DICT_VALUE	myval;
 
-	for (v = dictionary_values; v; v = v->next) {
-		if (v->attr == attr && v->value == val)
-			return v;
-	}
+	myval.attr = attr;
+	myval.value = val;
 
-	return NULL;
+	return rbtree_finddata(values_byvalue, &myval);
 }
 
 /*
@@ -704,24 +914,20 @@ DICT_VALUE * dict_valbyattr(int attr, int val)
  *      send it 0 as the attr. I hope this works the way it
  *      seems to. :) --kph
  */
-DICT_VALUE * dict_valbyname(int attr, const char *name)
+DICT_VALUE *dict_valbyname(int attr, const char *name)
 {
-	DICT_VALUE	*v;
+	DICT_VALUE	myval;
 
-	for (v = dictionary_values; v; v = v->next) {
-		if ((attr == 0 || v->attr == attr) &&
-		    strcasecmp(v->name, name) == 0)
-		 return v;
-               
-	}
+	myval.attr = attr;
+	strNcpy(myval.name, name, sizeof(myval.name));
 
-	return NULL;
+	return rbtree_finddata(values_byname, &myval);
 }
 
 /*
  *	Get the vendor PEC based on the vendor name
  */
-int dict_vendorname(const char *name)
+int dict_vendorbyname(const char *name)
 {
 	DICT_VENDOR *v;
 
@@ -729,10 +935,30 @@ int dict_vendorname(const char *name)
 	 *	Find the vendor, if any.
 	 */
 	for (v = dictionary_vendors; v; v = v->next) {
-		if (strcasecmp(name, v->vendorname) == 0) {
+		if (strcasecmp(name, v->name) == 0) {
 			return v->vendorpec;
 		}
 	}
 
 	return 0;
 }
+
+/*
+ *	Return the vendor struct based on the PEC.
+ */
+DICT_VENDOR *dict_vendorbyvalue(int vendor)
+{
+	DICT_VENDOR *v;
+
+	/*
+	 *	Find the vendor, if any.
+	 */
+	for (v = dictionary_vendors; v; v = v->next) {
+		if (vendor == v->vendorpec) {
+			return v;
+		}
+	}
+
+	return NULL;
+
+}
