@@ -1,7 +1,7 @@
 /*
  * valuepair.c	Functions to handle VALUE_PAIRs
  *
- * Version:	$Id: valuepair.c,v 1.62.2.2 2003/10/09 00:47:14 phampson Exp $
+ * Version:	$Id: valuepair.c,v 1.82 2004/05/06 14:37:21 aland Exp $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -20,7 +20,7 @@
  * Copyright 2000  The FreeRADIUS server project
  */
 
-static const char rcsid[] = "$Id: valuepair.c,v 1.62.2.2 2003/10/09 00:47:14 phampson Exp $";
+static const char rcsid[] = "$Id: valuepair.c,v 1.82 2004/05/06 14:37:21 aland Exp $";
 
 #include	"autoconf.h"
 
@@ -33,11 +33,11 @@ static const char rcsid[] = "$Id: valuepair.c,v 1.62.2.2 2003/10/09 00:47:14 pha
 
 #include	"libradius.h"
 
-#if HAVE_MALLOC_H
+#ifdef HAVE_MALLOC_H
 #  include	<malloc.h>
 #endif
 
-#if HAVE_REGEX_H
+#ifdef HAVE_REGEX_H
 #  include	<regex.h>
 #endif
 
@@ -75,8 +75,16 @@ VALUE_PAIR *paircreate(int attr, int type)
 	} else if (VENDOR(attr) == 0) {
 		sprintf(vp->name, "Attr-%u", attr);
 	} else {
-		sprintf(vp->name, "Vendor-%u-Attr-%u",
-			VENDOR(attr), attr & 0xffff);
+		DICT_VENDOR *v;
+
+		v = dict_vendorbyvalue(VENDOR(attr));
+		if (v) {
+			sprintf(vp->name, "%s-Attr-%u",
+				v->name, attr & 0xffff);
+		} else {
+			sprintf(vp->name, "Vendor-%u-Attr-%u",
+				VENDOR(attr), attr & 0xffff);
+		}
 	}
 	switch (vp->type) {
 		case PW_TYPE_INTEGER:
@@ -92,6 +100,17 @@ VALUE_PAIR *paircreate(int attr, int type)
 	return vp;
 }
 
+/*
+ *      release the memory used by a single attribute-value pair
+ *      just a wrapper around free() for now.
+ */
+void pairbasicfree(VALUE_PAIR *pair)
+{
+	/* clear the memory here */
+	memset(pair, 0, sizeof(*pair));
+	free(pair);
+}
+
 /*
  *	Release the memory used by a list of attribute-value
  *	pairs, and sets the pair pointer to NULL.
@@ -105,7 +124,7 @@ void pairfree(VALUE_PAIR **pair_ptr)
 
 	while (pair != NULL) {
 		next = pair->next;
-		free(pair);
+		pairbasicfree(pair);
 		pair = next;
 	}
 
@@ -136,7 +155,7 @@ void pairdelete(VALUE_PAIR **first, int attr)
 		next = i->next;
 		if (i->attribute == attr) {
 			*last = next;
-			free(i);
+			pairbasicfree(i);
 		} else {
 			last = &i->next;
 		}
@@ -159,6 +178,55 @@ void pairadd(VALUE_PAIR **first, VALUE_PAIR *add)
 	i->next = add;
 }
 
+/*
+ *	Add or replace a pair at the end of a VALUE_PAIR list.
+ */
+void pairreplace(VALUE_PAIR **first, VALUE_PAIR *replace)
+{
+	VALUE_PAIR *i, *next;
+	VALUE_PAIR **prev = first;
+
+	if (*first == NULL) {
+		*first = replace;
+		return;
+	}
+
+	/*
+	 *	Not an empty list, so find item if it is there, and
+	 *	replace it. Note, we always replace the first one, and
+	 *	we ignore any others that might exist.
+	 */
+	for(i = *first; i; i = next) {
+		next = i->next;
+
+		/*
+		 *	Found the first attribute, replace it,
+		 *	and return.
+		 */
+		if (i->attribute == replace->attribute) {
+			*prev = replace;
+
+			/*
+			 *	Should really assert that replace->next == NULL
+			 */
+			replace->next = next;
+			pairbasicfree(i);
+			return;
+		}
+
+		/*
+		 *	Point to where the attribute should go.
+		 */
+		prev = &i->next;
+	}
+
+	/*
+	 *	If we got here, we didn't find anything to replace, so
+	 *	stopped at the last item, which we just append to.
+	 */
+	*prev = replace;
+}
+
 /*
  *	Copy just a certain type of pairs.
  */
@@ -250,7 +318,7 @@ void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 		 */
 		if (i->attribute == PW_FALL_THROUGH ||
 		    (i->attribute != PW_HINT && i->attribute != PW_FRAMED_ROUTE)) {
-		  
+
 			found = pairfind(*to, i->attribute);
 			switch (i->operator) {
 
@@ -264,7 +332,7 @@ void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 					    (strcmp((char *)found->strvalue,
 						    (char *)i->strvalue) == 0)){
 						pairdelete(to, found->attribute);
-						
+
 						/*
 						 *	'tailto' may have been
 						 *	deleted...
@@ -278,9 +346,9 @@ void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 				tailfrom = i;
 				continue;
 				break;
-				
+
 /* really HAVE_REGEX_H */
-#if 0 
+#if 0
 				/*
 				 *  Attr-Name =~ "s/find/replace/"
 				 *
@@ -307,7 +375,7 @@ void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 			    q = strchr(str, *p);
 			    *(q++) = '\0';
 			    q[strlen(q) - 1] = '\0';
-			    
+
 			    regcomp(&reg, str, 0);
 			    if (regexec(&reg, found->strvalue,
 					1, match, 0) == 0) {
@@ -367,7 +435,7 @@ void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 			tailfrom->next = next;
 		else
 			*from = next;
-		
+
 		/*
 		 *	If ALL of the 'to' attributes have been deleted,
 		 *	then ensure that the 'tail' is updated to point
@@ -479,10 +547,20 @@ static int gettime(const char *valstr, time_t *lvalue)
 	struct tm	*tm, s_tm;
 	char		buf[64];
 	char		*p;
-	char		*f[3];
+	char		*f[4];
+	char            *tail = '\0';
 
-	time(&t);
-	tm = localtime_r(&t, &s_tm);
+	/*
+	 * Test for unix timestamp date
+	 */
+	*lvalue = strtoul(valstr, &tail, 10);
+	if (*tail == '\0') {
+		return 0;
+	}
+
+	tm = &s_tm;
+	memset(tm, 0, sizeof(*tm));
+	tm->tm_isdst = -1;	/* don't know, and don't care about DST */
 
 	strNcpy(buf, valstr, sizeof(buf));
 
@@ -490,6 +568,7 @@ static int gettime(const char *valstr, time_t *lvalue)
 	f[0] = mystrtok(&p, " \t");
 	f[1] = mystrtok(&p, " \t");
 	f[2] = mystrtok(&p, " \t");
+	f[3] = mystrtok(&p, " \t"); /* may, or may not, be present */
 	if (!f[0] || !f[1] || !f[2]) return -1;
 
 	/*
@@ -497,7 +576,7 @@ static int gettime(const char *valstr, time_t *lvalue)
 	 */
 	tm->tm_mon = 12;
 	for (i = 0; i < 3; i++) {
-		if (isalpha( (int) *f[i])) {	
+		if (isalpha( (int) *f[i])) {
 			/*
 			 *  Bubble the month to the front of the list
 			 */
@@ -548,6 +627,25 @@ static int gettime(const char *valstr, time_t *lvalue)
 		return -1;
 	}
 
+	/*
+	 *	There may be %H:%M:%S.  Parse it in a hacky way.
+	 */
+	if (f[3]) {
+		f[0] = f[3];	/* HH */
+		f[1] = strchr(f[0], ':'); /* find : separator */
+		if (!f[1]) return -1;
+
+		*(f[1]++) = '\0'; /* nuke it, and point to MM:SS */
+
+		f[2] = strchr(f[1], ':'); /* find : separator */
+		if (!f[2]) return -1;
+		*(f[2]++) = '\0';	/* nuke it, and point to SS */
+
+		tm->tm_hour = atoi(f[0]);
+		tm->tm_min = atoi(f[1]);
+		tm->tm_sec = atoi(f[2]);
+	}
+
 	/*
 	 *  Returns -1 on error.
 	 */
@@ -639,10 +737,11 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			/*
 			 *	Special case to convert filter to binary
 			 */
-		  	if ( filterBinary( vp, value ) < 0 ) {
-			  librad_log("failed to parse Ascend binary attribute: %s",
-				     librad_errstr);
-			  return NULL;
+			strNcpy(vp->strvalue, value, sizeof(vp->strvalue));
+		  	if (ascend_parse_filter(vp) < 0 ) {
+				librad_log("failed to parse Ascend binary attribute: %s",
+					   librad_errstr);
+				return NULL;
 			}
 			break;
 			/*
@@ -659,9 +758,20 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 				us = vp->strvalue;
 				vp->length = 0;
 
+
+				/*
+				 *	There is only one character,
+				 *	die.
+				 */
+				if ((strlen(cp) & 0x01) != 0) {
+					librad_log("Hex string is not an even length string.");
+					return NULL;
+				}
+				
+				
 				while (*cp && vp->length < MAX_STRING_LEN) {
 					unsigned int tmp;
-					
+
 					if (sscanf(cp, "%02x", &tmp) != 1) {
 						librad_log("Non-hex characters at %c%c", cp[0], cp[1]);
 						return NULL;
@@ -675,6 +785,25 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			}
 			break;
 
+		case PW_TYPE_IFID:
+			if (ifid_aton(value, vp->strvalue) == NULL) {
+				librad_log("failed to parse interface-id "
+					   "string \"%s\"", value);
+				return NULL;
+			}
+			vp->length = 8;
+			vp->strvalue[vp->length] = '\0';
+			break;
+
+		case PW_TYPE_IPV6ADDR:
+			if (ipv6_addr(value, vp->strvalue) < 0) {
+				librad_log("failed to parse IPv6 address "
+					   "string \"%s\"", value);
+				return NULL;
+			}
+			vp->length = 16; /* length of IPv6 address */
+			vp->strvalue[vp->length] = '\0';
+			break;
 			/*
 			 *  Anything else.
 			 */
@@ -686,6 +815,180 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 	return vp;
 }
 
+/*
+ *	Create a VALUE_PAIR from an ASCII attribute and value,
+ *	where the attribute name is in the form:
+ *
+ *	Attr-%d
+ *	Vendor-%d-Attr-%d
+ */
+static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
+				int operator)
+{
+	int		attr;
+	const char	*p;
+	VALUE_PAIR	*vp;
+	DICT_ATTR	*da;
+
+	/*
+	 *	Unknown attributes MUST be of type 'octets'
+	 */
+	if (value && (strncasecmp(value, "0x", 2) != 0)) {
+		goto error;
+	}
+
+	/*
+	 *	Attr-%d
+	 */
+	if (strncasecmp(attribute, "Attr-", 5) == 0) {
+		attr = atoi(attribute + 5);
+		p = attribute + 5;
+		p += strspn(p, "0123456789");
+		if (*p != 0) goto error;
+
+
+		/*
+		 *	Vendor-%d-Attr-%d
+		 */
+	} else if (strncasecmp(attribute, "Vendor-", 7) == 0) {
+		int vendor;
+
+		vendor = atoi(attribute + 7);
+		if ((vendor == 0) || (vendor > 65535)) goto error;
+
+		p = attribute + 7;
+		p += strspn(p, "0123456789");
+
+		/*
+		 *	Not Vendor-%d-Attr-%d
+		 */
+		if (strncasecmp(p, "-Attr-", 6) != 0) goto error;
+
+		p += 6;
+		attr = atoi(p);
+
+		p += strspn(p, "0123456789");
+		if (*p != 0) goto error;
+
+		if ((attr == 0) || (attr > 65535)) goto error;
+
+		attr |= (vendor << 16);
+
+		/*
+		 *	VendorName-Attr-%d
+		 */
+	} else if (((p = strchr(attribute, '-')) != NULL) &&
+		   (strncasecmp(p, "-Attr-", 6) == 0)) {
+		int vendor;
+		char buffer[256];
+
+		if ((p - attribute) >= sizeof(buffer)) goto error;
+
+		memcpy(buffer, attribute, p - attribute);
+		buffer[p - attribute] = '\0';
+
+		vendor = dict_vendorbyname(buffer);
+		if (vendor == 0) goto error;
+
+		p += 6;
+		attr = atoi(p);
+
+		p += strspn(p, "0123456789");
+		if (*p != 0) goto error;
+
+		if ((attr == 0) || (attr > 65535)) goto error;
+
+		attr |= (vendor << 16);
+
+	} else {		/* very much unknown: die */
+	error:
+		librad_log("Unknown attribute \"%s\"", attribute);
+		return NULL;
+	}
+
+	/*
+	 *	We've now parsed the attribute properly, and verified
+	 *	it to have value 'octets'.  Let's create it.
+	 */
+	if ((vp = (VALUE_PAIR *)malloc(sizeof(VALUE_PAIR))) == NULL) {
+		librad_log("out of memory");
+		return NULL;
+	}
+	memset(vp, 0, sizeof(VALUE_PAIR));
+	vp->type = PW_TYPE_OCTETS;
+
+	/*
+	 *	It may not be valid hex characters.  If not, die.
+	 */
+	if (pairparsevalue(vp, value) == NULL) {
+		pairfree(&vp);
+		return NULL;
+	}
+
+	/*
+	 *	Dictionary type over-rides what the caller says.
+	 *	This "converts" the parsed value into the appropriate
+	 *	type.
+	 *
+	 *	Also, normalize the name of the attribute...
+	 *
+	 *	Much of this code is copied from paircreate()
+	 */
+	if ((da = dict_attrbyvalue(attr)) != NULL) {
+		strcpy(vp->name, da->name);
+		vp->type = da->type;
+		vp->flags = da->flags;
+
+		/*
+		 *	Sanity check the type for length.  We don't
+		 *	want to look at attributes which are of the
+		 *	wrong length.
+		 */
+		switch (vp->type) {
+		case PW_TYPE_DATE:
+		case PW_TYPE_INTEGER:
+		case PW_TYPE_IPADDR: /* always kept in network byte order */
+			if (vp->length != 4) {
+			length_error:
+				pairfree(&vp);
+				librad_log("Attribute has invalid length");
+				return NULL;
+			}
+			memcpy(&vp->lvalue, vp->strvalue, sizeof(vp->lvalue));
+			break;
+
+		case PW_TYPE_IFID:
+			if (vp->length != 8) goto length_error;
+			break;
+
+		case PW_TYPE_IPV6ADDR:
+			if (vp->length != 16) goto length_error;
+			break;
+
+#ifdef ASCEND_BINARY
+		case PW_TYPE_ABINARY:
+			if (vp->length != 32) goto length_error;
+			break;
+#endif
+		default:	/* string, octets, etc. */
+			break;
+		}
+
+	} else if (VENDOR(attr) == 0) {
+		sprintf(vp->name, "Attr-%u", attr);
+	} else {
+		sprintf(vp->name, "Vendor-%u-Attr-%u",
+			VENDOR(attr), attr & 0xffff);
+	}
+
+	vp->attribute = attr;
+	vp->operator = (operator == 0) ? T_OP_EQ : operator;
+	vp->next = NULL;
+
+	return vp;
+}
+
+
 /*
  *	Create a VALUE_PAIR from an ASCII attribute and value.
  */
@@ -727,9 +1030,12 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 		 found_tag = 1;
 	}
 
+	/*
+	 *	It's not found in the dictionary, so we use
+	 *	another method to create the attribute.
+	 */
 	if ((da = dict_attrbyname(attribute)) == NULL) {
-		librad_log("Unknown attribute %s", attribute);
-		return NULL;
+		return pairmake_any(attribute, value, operator);
 	}
 
 	if ((vp = (VALUE_PAIR *)malloc(sizeof(VALUE_PAIR))) == NULL) {
@@ -753,7 +1059,7 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 	if (value && (*value == ':' && da->flags.has_tag)) {
 	        /* If we already found a tag, this is invalid */
 	        if(found_tag) {
-		        free(vp);
+		        pairbasicfree(vp);
 			librad_log("Duplicate tag %s for attribute %s",
 				   value, vp->name);
 			DEBUG("Duplicate tag %s for attribute %s\n",
@@ -774,8 +1080,8 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 		       else tag = 0;
 		}
 		found_tag = 1;
-	}	
-	
+	}
+
 	if (found_tag) {
 	  vp->flags.tag = tag;
 	}
@@ -819,23 +1125,23 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 		if (res != 0) {
 			char	msg[128];
 
-			regerror(res, &cre, msg, sizeof(msg));               
+			regerror(res, &cre, msg, sizeof(msg));
 			librad_log("Illegal regular expression in attribute: %s: %s",
 				vp->name, msg);
-			free(vp);
+			pairbasicfree(vp);
 			return NULL;
 		}
 		regfree(&cre);
 #else
 		librad_log("Regelar expressions not enabled in this build, error in attribute %s",
 				vp->name);
-		free(vp);
+		pairbasicfree(vp);
 		return NULL;
 #endif
 	}
 
 	if (value && (pairparsevalue(vp, value) == NULL)) {
-		free(vp);
+		pairbasicfree(vp);
 		return NULL;
 	}
 
@@ -855,7 +1161,7 @@ VALUE_PAIR *pairread(char **ptr, LRAD_TOKEN *eol)
 	LRAD_TOKEN	token, t, xlat;
 	VALUE_PAIR	*vp;
 
-	*eol = 0;
+	*eol = T_INVALID;
 
 	/* Get attribute. */
 	token = gettoken(ptr, attr, sizeof(attr));
@@ -911,19 +1217,43 @@ VALUE_PAIR *pairread(char **ptr, LRAD_TOKEN *eol)
 		vp = pairmake(attr, value, token);
 		break;
 
+		/*
+		 *	Perhaps do xlat's
+		 */
+	case T_DOUBLE_QUOTED_STRING:
+		p = strchr(value, '%');
+		if (p) {
+			vp = pairmake(attr, NULL, token);
+			if (!vp) {
+				*eol = T_INVALID;
+				return NULL;
+			}
+
+			strNcpy(vp->strvalue, value, sizeof(vp->strvalue));
+			vp->flags.do_xlat = 1;
+			vp->length = 0;
+		} else {
+			vp = pairmake(attr, value, token);
+		}
+		break;
+
+
 		/*
 		 *	Mark the pair to be allocated later.
 		 */
 	case T_BACK_QUOTED_STRING:
 		vp = pairmake(attr, NULL, token);
-		if (!vp) return vp;
-		
+		if (!vp) {
+			*eol = T_INVALID;
+			return NULL;
+		}
+
 		vp->flags.do_xlat = 1;
 		strNcpy(vp->strvalue, value, sizeof(vp->strvalue));
 		vp->length = 0;
 		break;
 	}
-	
+
 	return vp;
 }
 
@@ -948,7 +1278,7 @@ LRAD_TOKEN userparse(char *buffer, VALUE_PAIR **first_pair)
 	do {
 		previous_token = last_token;
 		if ((vp = pairread(&p, &last_token)) == NULL) {
-			return T_INVALID;
+			return last_token;
 		}
 		pairadd(first_pair, vp);
 	} while (*p && (last_token == T_COMMA));
@@ -966,3 +1296,60 @@ LRAD_TOKEN userparse(char *buffer, VALUE_PAIR **first_pair)
 	return last_token;
 }
 
+/*
+ *	Read valuepairs from the fp up to End-Of-File.
+ *
+ *	Hmm... this function is only used by radclient..
+ */
+VALUE_PAIR *readvp2(FILE *fp, int *pfiledone, const char *errprefix)
+{
+	char buf[8192];
+	LRAD_TOKEN last_token = T_EOL;
+	VALUE_PAIR *vp;
+	VALUE_PAIR *list;
+	int error = 0;
+
+	list = NULL;
+
+	while (!error && fgets(buf, sizeof(buf), fp) != NULL) {
+		/*
+		 *      If we get a '\n' by itself, we assume that's
+		 *      the end of that VP
+		 */
+		if ((buf[0] == '\n') && (list)) {
+			return list;
+		}
+		if ((buf[0] == '\n') && (!list)) {
+			continue;
+		}
+
+		/*
+		 *	Comments get ignored
+		 */
+		if (buf[0] == '#') continue;
+
+		/*
+		 *	Read all of the attributes on the current line.
+		 */
+		vp = NULL;
+		last_token = userparse(buf, &vp);
+		if (!vp) {
+			if (last_token != T_EOL) {
+				librad_perror(errprefix);
+				error = 1;
+				break;
+			}
+			break;
+		}
+
+		pairadd(&list, vp);
+		buf[0] = '\0';
+	}
+
+	if (error) pairfree(&list);
+
+	*pfiledone = 1;
+
+	return error ? NULL: list;
+}
+
