@@ -1,7 +1,7 @@
 /*
  * acct.c	Accounting routines.
  *
- * Version:	$Id: acct.c,v 1.23 2001/08/27 18:14:22 aland Exp $
+ * Version:	$Id: acct.c,v 1.30 2004/04/14 19:53:58 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  * Copyright 2000  Alan Curry <pacman@world.std.com>
  */
 
-static const char rcsid[] = "$Id: acct.c,v 1.23 2001/08/27 18:14:22 aland Exp $";
+static const char rcsid[] = "$Id: acct.c,v 1.30 2004/04/14 19:53:58 aland Exp $";
 
 #include "autoconf.h"
 #include "libradius.h"
@@ -50,20 +50,24 @@ int rad_accounting(REQUEST *request)
 		int		exec_wait;
 		VALUE_PAIR	*vp;
 		int		rcode;
+		int		acct_type = 0;
 
 		reply = module_preacct(request);
 		if (reply != RLM_MODULE_NOOP &&
 		    reply != RLM_MODULE_OK &&
 		    reply != RLM_MODULE_UPDATED)
 			return reply;
-		
+
 		/*
 		 *	Do accounting, ONLY the first time through.
 		 *	This is to ensure that we log the packet
 		 *	immediately, even if the proxy never does.
 		 */
-		reply = module_accounting(request);
-		
+		vp = pairfind(request->config_items, PW_ACCT_TYPE);
+		if (vp)
+			acct_type = vp->lvalue;
+		reply = module_accounting(acct_type,request);
+
 		/*
 		 *	See if we need to execute a program.
 		 *	FIXME: somehow cache this info, and only execute the
@@ -83,7 +87,7 @@ int rad_accounting(REQUEST *request)
 			exec_program = strdup((char *)vp->strvalue);
 			pairdelete(&request->reply->vps, PW_EXEC_PROGRAM_WAIT);
 		}
-		
+
 		/*
 		 *	If we want to exec a program, but wait for it,
 		 *	do it first before sending the reply, or
@@ -93,19 +97,32 @@ int rad_accounting(REQUEST *request)
 		 *	don't check the return code.
 		 */
 		if (exec_program) {
+			/*
+			 *	Wait for the answer.
+			 *	Don't look for a user message.
+			 *	Do look for returned VP's.
+			 */
 			rcode = radius_exec_program(exec_program, request,
-						    exec_wait, NULL);
+						    exec_wait, NULL, 0,
+						    request->packet->vps, &vp);
+			free(exec_program);
+
+			/*
+			 *	Always add the value-pairs to the reply.
+			 *
+			 *	If we're not waiting, then the pairs
+			 *	will be empty, so this won't matter.
+			 */
+			pairmove(&request->reply->vps, &vp);
+			pairfree(&vp);
+
 			if (exec_wait) {
 				if (rcode != 0) {
-					free(exec_program);
 					return reply;
 				}
 			}
 		}
 
-		if (exec_program) 
-			free(exec_program);
-
 		/*
 		 *	Maybe one of the preacct modules has decided
 		 *	that a proxy should be used. If so, get out of
@@ -121,9 +138,11 @@ int rad_accounting(REQUEST *request)
 	 *	We get here IF we're not proxying, OR if we've
 	 *	received the accounting reply from the end server,
 	 *	THEN we can reply to the NAS.
+	 *      If the accounting module returns NOOP, the data
+	 *      storage did not succeed, so radiusd should not send
+	 *      Accounting-Response.
 	 */
-	if (reply == RLM_MODULE_NOOP ||
-	    reply == RLM_MODULE_OK ||
+	if (reply == RLM_MODULE_OK ||
 	    reply == RLM_MODULE_UPDATED) {
 
 		/*
