@@ -1,7 +1,7 @@
 /*
  * auth.c	User authentication.
  *
- * Version:	$Id: auth.c,v 1.127 2003/06/24 14:22:19 pnixon Exp $
+ * Version:	$Id: auth.c,v 1.136 2004/03/08 21:51:03 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  * Copyright 2000  Jeff Carneal <jeff@apex.net>
  */
 
-static const char rcsid[] = "$Id: auth.c,v 1.127 2003/06/24 14:22:19 pnixon Exp $";
+static const char rcsid[] = "$Id: auth.c,v 1.136 2004/03/08 21:51:03 aland Exp $";
 
 #include "autoconf.h"
 #include "libradius.h"
@@ -31,7 +31,7 @@ static const char rcsid[] = "$Id: auth.c,v 1.127 2003/06/24 14:22:19 pnixon Exp
 #include <string.h>
 #include <ctype.h>
 
-#if HAVE_NETINET_IN_H
+#ifdef HAVE_NETINET_IN_H
 #	include <netinet/in.h>
 #endif
 
@@ -124,7 +124,7 @@ static int rad_authlog(const char *msg, REQUEST *request, int goodpass) {
 		username = request->username;
 	}
 
-	/* 
+	/*
 	 *	Clean up the username
 	 */
 	if (username == NULL) {
@@ -135,7 +135,7 @@ static int rad_authlog(const char *msg, REQUEST *request, int goodpass) {
 				clean_username, sizeof(clean_username));
 	}
 
-	/* 
+	/*
 	 *	Clean up the password
 	 */
 	if (mainconfig.log_auth_badpass || mainconfig.log_auth_goodpass) {
@@ -152,20 +152,20 @@ static int rad_authlog(const char *msg, REQUEST *request, int goodpass) {
 
 	if (goodpass) {
 		radlog(L_AUTH, "%s: [%s%s%s] (%s)",
-				msg,	
+				msg,
 				clean_username,
 				mainconfig.log_auth_goodpass ? "/" : "",
 				mainconfig.log_auth_goodpass ? clean_password : "",
 				auth_name(buf, sizeof(buf), request, 1));
 	} else {
 		radlog(L_AUTH, "%s: [%s%s%s] (%s)",
-				msg,	
+				msg,
 				clean_username,
 				mainconfig.log_auth_badpass ? "/" : "",
 				mainconfig.log_auth_badpass ? clean_password : "",
 				auth_name(buf, sizeof(buf), request, 1));
 	}
-	
+
 	return 0;
 }
 
@@ -212,14 +212,14 @@ int rad_check_password(REQUEST *request)
 	}
 
 	if (( auth_type_count > 1) && (debug_flag)) {
-		radlog(L_ERR, "Warning:  Found %d auth-types on request for user '%s'", 
+		radlog(L_ERR, "Warning:  Found %d auth-types on request for user '%s'",
 			auth_type_count, request->username->strvalue);
 	}
 
 	/*
-	 *	This means we have a proxy reply or an accept  
-	 *  and it wasn't rejected in the above loop.  So 
-	 *  that means it is accepted and we do no further 
+	 *	This means we have a proxy reply or an accept
+	 *  and it wasn't rejected in the above loop.  So
+	 *  that means it is accepted and we do no further
 	 *  authentication
 	 */
 	if ((auth_type == PW_AUTHTYPE_ACCEPT) || (request->proxy)) {
@@ -230,10 +230,15 @@ int rad_check_password(REQUEST *request)
 	/*
 	 *	Find the password from the users file.
 	 */
-	if ((password_pair = pairfind(request->config_items, PW_CRYPT_PASSWORD)) != NULL)
-		auth_type = PW_AUTHTYPE_CRYPT;
-	else
+	if ((password_pair = pairfind(request->config_items, PW_CRYPT_PASSWORD)) != NULL) {
+		/*
+		 *	Re-write Auth-Type, but ONLY if it isn't already
+		 *	set.
+		 */
+		if (auth_type == -1) auth_type = PW_AUTHTYPE_CRYPT;
+	} else {
 		password_pair = pairfind(request->config_items, PW_PASSWORD);
+	}
 
 	if (auth_type < 0) {
 		if (password_pair) {
@@ -270,7 +275,7 @@ int rad_check_password(REQUEST *request)
 					"(No Crypt-Password configured for the user)", request, 0);
 				return -1;
 			}
-					
+
 			switch (lrad_crypt_check((char *)auth_item->strvalue,
 									 (char *)password_pair->strvalue)) {
 			case -1:
@@ -375,6 +380,53 @@ int rad_check_password(REQUEST *request)
 	return result;
 }
 
+/*
+ *	Post-authentication step processes the response before it is
+ *	sent to the NAS. It can receive both Access-Accept and Access-Reject
+ *	replies.
+ */
+int rad_postauth(REQUEST *request)
+{
+	int	result;
+	int	postauth_type = 0;
+	VALUE_PAIR	*postauth_type_item = NULL;
+
+	/*
+	 *	Do post-authentication calls. ignoring the return code.
+	 */
+	postauth_type_item = pairfind(request->config_items, PW_POST_AUTH_TYPE);
+	if (postauth_type_item)
+		postauth_type = postauth_type_item->lvalue;
+	result = module_post_auth(postauth_type, request);
+	switch (result) {
+	default:
+	  break;
+
+	  /*
+	   *	The module failed, or said to reject the user: Do so.
+	   */
+	case RLM_MODULE_FAIL:
+	case RLM_MODULE_REJECT:
+	case RLM_MODULE_USERLOCK:
+	case RLM_MODULE_INVALID:
+	  request->reply->code = PW_AUTHENTICATION_REJECT;
+	  result = RLM_MODULE_REJECT;
+	  break;
+
+	  /*
+	   *	The module had a number of OK return codes.
+	   */
+	case RLM_MODULE_NOTFOUND:
+	case RLM_MODULE_NOOP:
+	case RLM_MODULE_UPDATED:
+	case RLM_MODULE_OK:
+	case RLM_MODULE_HANDLED:
+	  result = RLM_MODULE_OK;
+	  break;
+	}
+	return result;
+}
+
 /*
  *	Process and reply to an authentication request
  *
@@ -389,8 +441,6 @@ int rad_authenticate(REQUEST *request)
 	VALUE_PAIR	*auth_item;
 	VALUE_PAIR	*module_msg;
 	VALUE_PAIR	*tmp = NULL;
-	VALUE_PAIR	*autz_type_item = NULL;
-	VALUE_PAIR	*postauth_type_item = NULL;
 	int		result, r;
 	char		umsg[MAX_STRING_LEN + 1];
 	const char	*user_msg = NULL;
@@ -401,7 +451,6 @@ int rad_authenticate(REQUEST *request)
 	char		buf[1024], logstr[1024];
 	char		autz_retry = 0;
 	int		autz_type = 0;
-	int		postauth_type = 0;
 
 	password = "";
 
@@ -411,6 +460,9 @@ int rad_authenticate(REQUEST *request)
 	 *	to add an initial Auth-Type: Auth-Accept for success,
 	 *	Auth-Reject for fail. We also need to add the reply
 	 *	pairs from the server to the initial reply.
+	 *
+	 *	Huh?  If the request wasn't an authentication request,
+	 *	WTF are we doing here?
 	 */
 	if ((request->proxy_reply) &&
 	    (request->packet->code == PW_AUTHENTICATION_REQUEST)) {
@@ -468,7 +520,7 @@ int rad_authenticate(REQUEST *request)
 		request->password = pairfind(request->packet->vps,
 					     PW_PASSWORD);
 	}
-	  
+
 	/*
 	 *	Discover which password we want to use.
 	 */
@@ -480,10 +532,10 @@ int rad_authenticate(REQUEST *request)
 		/*
 		 *	Maybe there's a CHAP-Password?
 		 */
-		if ((auth_item = pairfind(request->packet->vps, 
+		if ((auth_item = pairfind(request->packet->vps,
 				PW_CHAP_PASSWORD)) != NULL) {
 			password = "<CHAP-PASSWORD>";
-		
+
 		} else {
 			/*
 			 *	No password we recognize.
@@ -492,7 +544,7 @@ int rad_authenticate(REQUEST *request)
 		}
 	}
 	request->password = auth_item;
-	
+
 	/*
 	 *	Get the user's authorization information from the database
 	 */
@@ -514,14 +566,10 @@ autz_redo:
 			}
 			request->reply->code = PW_AUTHENTICATION_REJECT;
 		}
-
-		/*
-		 *	Hope that the module returning REJECT is smart
-		 *	enough to do pairfre(&request->reply->vps)...
-		 */
 		return r;
 	}
 	if (!autz_retry){
+		VALUE_PAIR	*autz_type_item = NULL;
 		autz_type_item = pairfind(request->config_items, PW_AUTZ_TYPE);
 		if (autz_type_item){
 			autz_type = autz_type_item->lvalue;
@@ -530,7 +578,6 @@ autz_redo:
 		}
 	}
 
-
 	/*
 	 *	If we haven't already proxied the packet, then check
 	 *	to see if we should.  Maybe one of the authorize
@@ -538,8 +585,23 @@ autz_redo:
 	 *	so, get out of here and send the packet.
 	 */
 	if ((request->proxy == NULL) &&
-			(pairfind(request->config_items, PW_PROXY_TO_REALM) != NULL)) {
-		return RLM_MODULE_OK;
+	    ((tmp = pairfind(request->config_items, PW_PROXY_TO_REALM)) != NULL)) {
+		REALM *realm;
+
+		/*
+		 *	Catch users who set Proxy-To-Realm to a LOCAL
+		 *	realm (sigh).
+		 */
+		realm = realm_find(tmp->strvalue, 0);
+		if (realm && (realm->ipaddr == htonl(INADDR_NONE))) {
+			DEBUG2("  WARNING: You set Proxy-To-Realm = %s, but it is a LOCAL realm!  Cancelling invalid proxy request.", realm->realm);
+		} else {
+			/*
+			 *	Don't authenticate, as the request is
+			 *	proxied.
+			 */
+			return RLM_MODULE_OK;
+		}
 	}
 
 	/*
@@ -559,7 +621,7 @@ autz_redo:
 			return RLM_MODULE_HANDLED;
 		}
 	} while(0);
-	
+
 	/*
 	 *	Failed to validate the user.
 	 *
@@ -598,7 +660,7 @@ autz_redo:
 	}
 
 	if (result >= 0 &&
-			(check_item = pairfind(request->config_items, PW_SIMULTANEOUS_USE)) != NULL) {
+	    (check_item = pairfind(request->config_items, PW_SIMULTANEOUS_USE)) != NULL) {
 		VALUE_PAIR	*session_type;
 		int		sess_type = 0;
 
@@ -626,7 +688,7 @@ autz_redo:
 			}
 			if (!mpp_ok){
 				if (check_item->lvalue > 1) {
-		  		snprintf(umsg, sizeof(umsg), 
+		  		snprintf(umsg, sizeof(umsg),
 							"\r\nYou are already logged in %d times  - access denied\r\n\n",
 							(int)check_item->lvalue);
 					user_msg = umsg;
@@ -663,12 +725,18 @@ autz_redo:
 		 */
 		r = timestr_match((char *)check_item->strvalue,
 				  request->timestamp);
-		/*
-		 *	Session-Timeout needs to be at least
-		 *	60 seconds, some terminal servers
-		 *	ignore smaller values.
-		 */
-		if (r < 60) {
+
+		if (r == 0) {	/* unlimited */
+			/*
+			 *	Do nothing: login-time is OK.
+			 */
+
+			/*
+			 *	Session-Timeout needs to be at least
+			 *	60 seconds, some terminal servers
+			 *	ignore smaller values.
+			 */
+		} else if (r < 60) {
 			/*
 			 *	User called outside allowed time interval.
 			 */
@@ -681,7 +749,7 @@ autz_redo:
 			tmp = pairmake("Reply-Message", user_msg, T_OP_SET);
 			request->reply->vps = tmp;
 
-			snprintf(logstr, sizeof(logstr), "Outside allowed timespan (time allowed %s)", 
+			snprintf(logstr, sizeof(logstr), "Outside allowed timespan (time allowed %s)",
 				 check_item->strvalue);
 			rad_authlog(logstr, request, 1);
 
@@ -708,9 +776,22 @@ autz_redo:
 	}
 
 	/*
-	 *	Result should be >= 0 here - if not, we return.
+	 *	Result should be >= 0 here - if not, it means the user
+	 *	is rejected, so we overwrite the Post-Auth-Type with
+	 *	the value REJECT and call the post-authentication
+	 *	step.
 	 */
 	if (result < 0) {
+		DICT_VALUE *dval;
+
+		dval = dict_valbyname(PW_POST_AUTH_TYPE, "REJECT");
+		if (dval) {
+			pairdelete(&request->config_items, PW_POST_AUTH_TYPE);
+			tmp = paircreate(PW_POST_AUTH_TYPE, PW_TYPE_INTEGER);
+			tmp->lvalue = dval->value;
+			pairadd(&request->config_items, tmp);
+			rad_postauth(request);
+		}
 		return RLM_MODULE_OK;
 	}
 
@@ -727,17 +808,18 @@ autz_redo:
 	 *	Add the port number to the Framed-IP-Address if
 	 *	vp->addport is set.
 	 */
-	if (((tmp = pairfind(request->reply->vps, 
+	if (((tmp = pairfind(request->reply->vps,
 			     PW_FRAMED_IP_ADDRESS)) != NULL) &&
 	    (tmp->flags.addport != 0)) {
 		VALUE_PAIR *vpPortId;
-		
+
 		/*
 		 *  Find the NAS port ID.
 		 */
 		if ((vpPortId = pairfind(request->packet->vps,
 					 PW_NAS_PORT)) != NULL) {
-		  tmp->lvalue = htonl(ntohl(tmp->lvalue) + vpPortId->lvalue);
+		  unsigned long tvalue = ntohl(tmp->lvalue);
+		  tmp->lvalue = htonl(tvalue + vpPortId->lvalue);
 		  tmp->flags.addport = 0;
 		  ip_ntoa(tmp->strvalue, tmp->lvalue);
 		} else {
@@ -814,7 +896,7 @@ autz_redo:
 			tmp = pairmake("Reply-Message", user_msg, T_OP_SET);
 
 			pairadd(&request->reply->vps, tmp);
-			rad_authlog("Login incorrect (external check failed)", 
+			rad_authlog("Login incorrect (external check failed)",
 					request, 0);
 
 			return RLM_MODULE_REJECT;
@@ -885,43 +967,10 @@ autz_redo:
 				    NULL, 0, request->packet->vps, NULL);
 	}
 
-	if (exec_program) 
+	if (exec_program)
 		free(exec_program);
 
-	/*
-	 *	Do post-authentication calls. ignoring the return code.
-	 *	If the post-authentication
-	 */
-	postauth_type_item = pairfind(request->config_items, PW_POST_AUTH_TYPE);
-	if (postauth_type_item)
-		postauth_type = postauth_type_item->lvalue;
-	result = module_post_auth(postauth_type, request);
-	switch (result) {
-	default:
-	  break;
-	  
-	  /*
-	   *	The module failed, or said to reject the user: Do so.
-	   */
-	case RLM_MODULE_FAIL:
-	case RLM_MODULE_REJECT:
-	case RLM_MODULE_USERLOCK:
-	case RLM_MODULE_INVALID:
-	  request->reply->code = PW_AUTHENTICATION_REJECT;
-	  result = RLM_MODULE_REJECT;
-	  break;
-
-	  /*
-	   *	The module had a number of OK return codes.
-	   */
-	case RLM_MODULE_NOTFOUND:
-	case RLM_MODULE_NOOP:
-	case RLM_MODULE_UPDATED:
-	case RLM_MODULE_OK:
-	case RLM_MODULE_HANDLED:
-	  result = RLM_MODULE_OK;
-	  break;
-	}
+	result = rad_postauth(request);
 
 	return result;
 }
