@@ -1,7 +1,7 @@
 /*
  * mainconf.c	Handle the server's configuration.
  *
- * Version:	$Id: mainconfig.c,v 1.20.2.1 2003/07/22 18:17:27 aland Exp $
+ * Version:	$Id: mainconfig.c,v 1.36.2.2 2004/06/08 15:09:40 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -31,6 +31,10 @@
 #include <netinet/in.h>
 #endif
 
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
+
 #include "radiusd.h"
 #include "rad_assert.h"
 #include "conffile.h"
@@ -39,6 +43,8 @@
 #include <sys/resource.h>
 #include <unistd.h>
 #include <sys/types.h>
+#include <sys/socket.h>
+#include <netdb.h>
 #include <sys/stat.h>
 #include <grp.h>
 #include <pwd.h>
@@ -57,6 +63,7 @@ static gid_t server_gid;
  */
 static const char *localstatedir = NULL;
 static const char *prefix = NULL;
+static int auth_port = 0;
 
 /*
  *  Map the proxy server configuration parameters to variables.
@@ -94,7 +101,7 @@ static CONF_PARSER server_config[] = {
 	 *	files.
 	 */
 	{ "prefix",             PW_TYPE_STRING_PTR, 0, &prefix,            "/usr/local"},
-	{ "localstatedir",      PW_TYPE_STRING_PTR, 0, &localstatedir,     "${prefix}/var"}, 
+	{ "localstatedir",      PW_TYPE_STRING_PTR, 0, &localstatedir,     "${prefix}/var"},
 	{ "logdir",             PW_TYPE_STRING_PTR, 0, &radlog_dir,        "${localstatedir}/log"},
 	{ "libdir",             PW_TYPE_STRING_PTR, 0, &radlib_dir,        "${prefix}/lib"},
 	{ "radacctdir",         PW_TYPE_STRING_PTR, 0, &radacct_dir,       "${logdir}/radacct" },
@@ -134,7 +141,8 @@ static CONF_PARSER server_config[] = {
  *	Xlat for %{config:section.subsection.attribute}
  */
 static int xlat_config(void *instance, REQUEST *request,
-		       char *fmt, char *out, int outlen,
+		       char *fmt, char *out,
+		       size_t outlen,
 		       RADIUS_ESCAPE_STRING func)
 {
 	CONF_SECTION *cs;
@@ -199,7 +207,7 @@ static int xlat_config(void *instance, REQUEST *request,
 			outlen = strlen(value) + 1;
 		}
 	}
-	
+
 	return func(out, outlen, value);
 }
 
@@ -207,7 +215,8 @@ static int xlat_config(void *instance, REQUEST *request,
 /*
  *	Recursively make directories.
  */
-static int r_mkdir(const char *part) {
+static int r_mkdir(const char *part)
+{
 	char *ptr, parentdir[500];
 	struct stat st;
 
@@ -229,15 +238,14 @@ static int r_mkdir(const char *part) {
 		return(1);
 	}
 
-	fprintf(stderr, "Created directory %s\n", part);
-
 	return(0);
 }
-		
+
 /*
  *	Checks if the log directory is writeable by a particular user.
  */
-static int radlogdir_iswritable(const char *effectiveuser) {
+static int radlogdir_iswritable(const char *effectiveuser)
+{
 	struct passwd *pwent;
 
 	if (radlog_dir[0] != '/')
@@ -250,11 +258,11 @@ static int radlogdir_iswritable(const char *effectiveuser) {
 	if (strstr(radlog_dir, "radius") == NULL)
 		return(0);
 
-	/* we have a logdir that mentions 'radius', so it's probably 
-	 * safe to chown the immediate directory to be owned by the normal 
+	/* we have a logdir that mentions 'radius', so it's probably
+	 * safe to chown the immediate directory to be owned by the normal
 	 * process owner. we gotta do it before we give up root.  -chad
 	 */
-	
+
 	if (!effectiveuser) {
 		return 1;
 	}
@@ -271,12 +279,11 @@ static int radlogdir_iswritable(const char *effectiveuser) {
 }
 
 
-static int switch_users(void) {
-	
-	/*
-	 *  Switch UID and GID to what is specified in the config file
-	 */
-
+/*
+ *  Switch UID and GID to what is specified in the config file
+ */
+static int switch_users(void)
+{
 	/*  Set GID.  */
 	if (mainconfig.gid_name != NULL) {
 		struct group *gr;
@@ -292,9 +299,12 @@ static int switch_users(void) {
 		}
 		server_gid = gr->gr_gid;
 		if (setgid(server_gid) < 0) {
-			radlog(L_ERR|L_CONS, "Failed setting Group to %s: %s", mainconfig.gid_name, strerror(errno));
+			radlog(L_ERR|L_CONS, "Failed setting Group to %s: %s",
+			       mainconfig.gid_name, strerror(errno));
 			exit(1);
 		}
+	} else {
+		server_gid = getgid();
 	}
 
 	/*  Set UID.  */
@@ -311,6 +321,14 @@ static int switch_users(void) {
 			exit(1);
 		}
 		server_uid = pw->pw_uid;
+#ifdef HAVE_INITGROUPS
+		if (initgroups(mainconfig.uid_name, server_gid) < 0) {
+			if (errno != EPERM) {
+				radlog(L_ERR|L_CONS, "Failed setting supplementary groups for User %s: %s", mainconfig.uid_name, strerror(errno));
+				exit(1);
+			}
+		}
+#endif
 		if (setuid(server_uid) < 0) {
 			radlog(L_ERR|L_CONS, "Failed setting User to %s: %s", mainconfig.uid_name, strerror(errno));
 			exit(1);
@@ -320,7 +338,7 @@ static int switch_users(void) {
 }
 
 
-/* 
+/*
  * Create the linked list of realms from the new configuration type
  * This way we don't have to change to much in the other source-files
  */
@@ -333,11 +351,13 @@ static int generate_realms(const char *filename)
 	char *name2;
 
 	tail = &my_realms;
-	for (cs = cf_subsection_find_next(mainconfig.config, NULL, "realm"); cs != NULL;
-			cs = cf_subsection_find_next(mainconfig.config, cs, "realm")) {
+	for (cs = cf_subsection_find_next(mainconfig.config, NULL, "realm");
+	     cs != NULL;
+	     cs = cf_subsection_find_next(mainconfig.config, cs, "realm")) {
 		name2 = cf_section_name2(cs);
 		if (!name2) {
-			radlog(L_CONS|L_ERR, "%s[%d]: Missing realm name", filename, cf_section_lineno(cs));
+			radlog(L_CONS|L_ERR, "%s[%d]: Missing realm name",
+			       filename, cf_section_lineno(cs));
 			return -1;
 		}
 		/*
@@ -370,15 +390,22 @@ static int generate_realms(const char *filename)
 				c->auth_port = auth_port;
 			} else {
 				c->ipaddr = ip_getaddr(authhost);
+				if (c->ipaddr == htonl(INADDR_NONE)) {
+					radlog(L_ERR, "%s[%d]: Host %s not found",
+					       filename, cf_section_lineno(cs),
+					       authhost);
+					return -1;
+				}
 			}
 
-			/* 
+			/*
 			 * Double check length, just to be sure!
 			 */
 			if (strlen(authhost) >= sizeof(c->server)) {
 				radlog(L_ERR, "%s[%d]: Server name of length %d is greater than allowed: %d",
 				       filename, cf_section_lineno(cs),
-				       strlen(authhost), sizeof(c->server) - 1);
+				       (int) strlen(authhost),
+				       (int) sizeof(c->server) - 1);
 				return -1;
 			}
 		}
@@ -388,13 +415,13 @@ static int generate_realms(const char *filename)
 		 */
 		if ((accthost = cf_section_value_find(cs, "accthost")) == NULL) {
 			c->acct_ipaddr = htonl(INADDR_NONE);
-			c->acct_port = acct_port;
+			c->acct_port = 0;
 		} else {
 			if ((s = strchr(accthost, ':')) != NULL) {
 				*s++ = 0;
-				c->acct_port = atoi(s);	
+				c->acct_port = atoi(s);
 			} else {
-				c->acct_port = acct_port;
+				c->acct_port = auth_port + 1;
 			}
 			if (strcmp(accthost, "LOCAL") == 0) {
 				/*
@@ -402,29 +429,37 @@ static int generate_realms(const char *filename)
 				 *	secret, or port.
 				 */
 				c->acct_ipaddr = htonl(INADDR_NONE);
-				c->acct_port = acct_port;
+				c->acct_port = 0;
 			} else {
 				c->acct_ipaddr = ip_getaddr(accthost);
+				if (c->acct_ipaddr == htonl(INADDR_NONE)) {
+					radlog(L_ERR, "%s[%d]: Host %s not found",
+					       filename, cf_section_lineno(cs),
+					       accthost);
+					return -1;
+				}
 			}
 
 			if (strlen(accthost) >= sizeof(c->acct_server)) {
 				radlog(L_ERR, "%s[%d]: Server name of length %d is greater than allowed: %d",
 				       filename, cf_section_lineno(cs),
-				       strlen(accthost), sizeof(c->acct_server) - 1);
+				       (int) strlen(accthost),
+				       (int) sizeof(c->acct_server) - 1);
 				return -1;
 			}
 		}
 
 		if (strlen(name2) >= sizeof(c->realm)) {
 			radlog(L_ERR, "%s[%d]: Realm name of length %d is greater than allowed %d",
-					filename, cf_section_lineno(cs),
-					strlen(name2), sizeof(c->server) - 1);
+			       filename, cf_section_lineno(cs),
+			       (int) strlen(name2),
+			       (int) sizeof(c->server) - 1);
 			return -1;
 		}
-		
+
 		strcpy(c->realm, name2);
                 if (authhost) strcpy(c->server, authhost);
-		if (accthost) strcpy(c->acct_server, accthost);	
+		if (accthost) strcpy(c->acct_server, accthost);
 
 		/*
 		 *	If one or the other of authentication/accounting
@@ -438,7 +473,7 @@ static int generate_realms(const char *filename)
 				       filename, cf_section_lineno(cs), name2);
 				return -1;
 			}
-			
+
 			if (strlen(s) >= sizeof(c->secret)) {
 				radlog(L_ERR, "%s[%d]: Secret of length %d is greater than the allowed maximum of %d.",
 				       filename, cf_section_lineno(cs),
@@ -449,7 +484,7 @@ static int generate_realms(const char *filename)
 		}
 
 		c->striprealm = 1;
-		
+
 		if ((cf_section_value_find(cs, "nostrip")) != NULL)
 			c->striprealm = 0;
 		if ((cf_section_value_find(cs, "noacct")) != NULL)
@@ -544,23 +579,25 @@ static int generate_realms(const char *filename)
  *	type.  This way we don't have to change too much in the other
  *	source-files.
  */
-static int generate_clients(const char *filename)
+static RADCLIENT *generate_clients(const char *filename, CONF_SECTION *section)
 {
 	CONF_SECTION	*cs;
-	RADCLIENT	*c;
+	RADCLIENT	*list, *c;
 	char		*hostnm, *secret, *shortnm, *netmask;
 	char            *nastype, *login, *password;
 	char		*name2;
 
-	for (cs = cf_subsection_find_next(mainconfig.config, NULL, "client");
-	     cs != NULL; 
-	     cs = cf_subsection_find_next(mainconfig.config, cs, "client")) {
+	list = NULL;
+	for (cs = cf_subsection_find_next(section, NULL, "client");
+	     cs != NULL;
+	     cs = cf_subsection_find_next(section, cs, "client")) {
 
 		name2 = cf_section_name2(cs);
 		if (!name2) {
 			radlog(L_CONS|L_ERR, "%s[%d]: Missing client name",
 			       filename, cf_section_lineno(cs));
-			return -1;
+			clients_free(list);
+			return NULL;
 		}
 		/*
 		 * Check the lengths, we don't want any core dumps
@@ -570,13 +607,15 @@ static int generate_clients(const char *filename)
 		if((secret = cf_section_value_find(cs, "secret")) == NULL) {
 			radlog(L_ERR, "%s[%d]: Missing secret for client: %s",
 				filename, cf_section_lineno(cs), name2);
-			return -1;
+			clients_free(list);
+			return NULL;
 		}
 
 		if((shortnm = cf_section_value_find(cs, "shortname")) == NULL) {
 			radlog(L_ERR, "%s[%d]: Missing shortname for client: %s",
 				filename, cf_section_lineno(cs), name2);
-			return -1;
+			clients_free(list);
+			return NULL;
 		}
 
 		netmask = strchr(hostnm, '/');
@@ -585,39 +624,45 @@ static int generate_clients(const char *filename)
 			radlog(L_ERR, "%s[%d]: Secret of length %d is greater than the allowed maximum of %d.",
 				filename, cf_section_lineno(cs),
 				strlen(secret), sizeof(c->secret) - 1);
-			return -1;
+			clients_free(list);
+			return NULL;
 		}
+
 		if (strlen(shortnm) > sizeof(c->shortname)) {
 			radlog(L_ERR, "%s[%d]: Client short name of length %d is greater than the allowed maximum of %d.",
 					filename, cf_section_lineno(cs),
-					strlen(shortnm), sizeof(c->shortname) - 1);
-			return -1;
+			       strlen(shortnm), sizeof(c->shortname) - 1);
+			clients_free(list);
+			return NULL;
 		}
 
 		if((nastype = cf_section_value_find(cs, "nastype")) != NULL) {
-		        if(strlen(nastype) >= sizeof(c->nastype)) { 
+		        if(strlen(nastype) >= sizeof(c->nastype)) {
 			       radlog(L_ERR, "%s[%d]: nastype of length %d longer than the allowed maximum of %d",
-				      filename, cf_section_lineno(cs), 
+				      filename, cf_section_lineno(cs),
 				      strlen(nastype), sizeof(c->nastype) - 1);
-			       return -1;
+			       clients_free(list);
+			       return NULL;
 			}
 		}
 
 		if((login = cf_section_value_find(cs, "login")) != NULL) {
-		        if(strlen(login) >= sizeof(c->login)) { 
+		        if(strlen(login) >= sizeof(c->login)) {
 			       radlog(L_ERR, "%s[%d]: login of length %d longer than the allowed maximum of %d",
-				      filename, cf_section_lineno(cs), 
+				      filename, cf_section_lineno(cs),
 				      strlen(login), sizeof(c->login) - 1);
-			       return -1;
+			       clients_free(list);
+			       return NULL;
 			}
 		}
 
 		if((password = cf_section_value_find(cs, "password")) != NULL) {
-		        if(strlen(password) >= sizeof(c->password)) { 
+		        if(strlen(password) >= sizeof(c->password)) {
 			       radlog(L_ERR, "%s[%d]: password of length %d longer than the allowed maximum of %d",
-				      filename, cf_section_lineno(cs), 
+				      filename, cf_section_lineno(cs),
 				      strlen(password), sizeof(c->password) - 1);
-			       return -1;
+			       clients_free(list);
+			       return NULL;
 			}
 		}
 
@@ -632,18 +677,20 @@ static int generate_clients(const char *filename)
 		 */
 		c->netmask = ~0;
 		if (netmask) {
-			int i, mask_length;
+			int mask_length;
 
 			mask_length = atoi(netmask + 1);
 			if ((mask_length < 0) || (mask_length > 32)) {
 				radlog(L_ERR, "%s[%d]: Invalid value '%s' for IP network mask.",
 						filename, cf_section_lineno(cs), netmask + 1);
-				return -1;
+				clients_free(list);
+				return NULL;
 			}
-			
-			c->netmask = (1 << 31);
-			for (i = 1; i < mask_length; i++) {
-				c->netmask |= (c->netmask >> 1);
+
+			if (mask_length == 0) {
+				c->netmask = 0;
+			} else {
+				c->netmask = ~0 << (32 - mask_length);
 			}
 
 			*netmask = '\0';
@@ -654,7 +701,8 @@ static int generate_clients(const char *filename)
 		if (c->ipaddr == INADDR_NONE) {
 			radlog(L_CONS|L_ERR, "%s[%d]: Failed to look up hostname %s",
 					filename, cf_section_lineno(cs), hostnm);
-			return -1;
+			clients_free(list);
+			return NULL;
 		}
 
 		/*
@@ -678,13 +726,392 @@ static int generate_clients(const char *filename)
 		if(password != NULL)
 		        strcpy(c->password, password);
 
-		c->next =mainconfig. clients;
-		mainconfig.clients = c;
+		c->next = list;
+		list = c;
+	}
+
+	return list;
+}
+
+
+/*
+ *	Code for handling listening on multiple ports.
+ */
+static rad_listen_t listen_inst;
+static const char *listen_type = NULL;
+
+static const CONF_PARSER listen_config[] = {
+	{ "ipaddr", PW_TYPE_IPADDR,
+	  offsetof(rad_listen_t,ipaddr), NULL, "0.0.0.0" },
+
+	{ "port", PW_TYPE_INTEGER,
+	  offsetof(rad_listen_t,port), NULL, "0" },
+
+	{ "type", PW_TYPE_STRING_PTR,
+	  0, &listen_type, "" },
+
+	{ NULL, -1, 0, NULL, NULL }		/* end the list */
+};
+
+static const LRAD_NAME_NUMBER listen_compare[] = {
+	{ "auth",	RAD_LISTEN_AUTH },
+	{ "acct",	RAD_LISTEN_ACCT },
+	{ NULL, 0 },
+};
+
+
+/*
+ *	Free a linked list of listeners;
+ */
+static void listen_free(rad_listen_t *list)
+{
+	while (list) {
+		rad_listen_t *next = list->next;
+		
+		/*
+		 *	The code below may have eaten the FD.
+		 */
+		if (list->fd >= 0) close(list->fd);
+		free(list);
+		
+		list = next;
+	}
+}
+
+/*
+ *	Binds a listener to a socket.
+ */
+static int listen_bind(rad_listen_t *this)
+{
+	struct sockaddr salocal;
+	struct sockaddr_in *sa;
+
+	rad_listen_t	**last;
+
+	/*
+	 *	If the port is zero, then it means the appropriate
+	 *	thing from /etc/services.
+	 */
+	if (this->port == 0) {
+		struct servent	*svp;
+
+		switch (this->type) {
+		case RAD_LISTEN_AUTH:
+			svp = getservbyname ("radius", "udp");
+			if (svp != NULL) {
+				this->port = ntohs(svp->s_port);
+			} else {
+				this->port = PW_AUTH_UDP_PORT;
+			}
+			break;
+
+		case RAD_LISTEN_ACCT:
+			svp = getservbyname ("radacct", "udp");
+			if (svp != NULL) {
+				this->port = ntohs(svp->s_port);
+			} else {
+				this->port = PW_ACCT_UDP_PORT;
+			}
+			break;
+
+		default:
+			radlog(L_ERR|L_CONS, "ERROR: Non-fatal internal sanity check failed in bind.");
+			return -1;
+		}
+	}
+
+	/*
+	 *	Find it in the old list, AFTER updating the port.  If
+	 *	it's there, use that, rather than creating a new
+	 *	socket.  This allows HUP's to re-use the old sockets,
+	 *	which means that packets waiting in the socket queue
+	 *	don't get lost.  */
+	for (last = &mainconfig.listen;
+	     *last != NULL;
+	     last = &((*last)->next)) {
+		if ((this->ipaddr == (*last)->ipaddr) &&
+		    (this->type == (*last)->type) &&
+		    (this->port == (*last)->port)) {
+			this->fd = (*last)->fd;
+			(*last)->fd = -1;
+			return 0;
+		}
+	}
+
+	/*
+	 *	Create the socket.
+	 */
+	this->fd = socket(AF_INET, SOCK_DGRAM, 0);
+	if (this->fd < 0) {
+		return -1;
+	}
+	
+	sa = (struct sockaddr_in *) &salocal;
+	memset ((char *) sa, '\0', sizeof(salocal));
+	sa->sin_family = AF_INET;
+	sa->sin_addr.s_addr = this->ipaddr;
+	sa->sin_port = htons(this->port);
+	
+	if (bind(this->fd, &salocal, sizeof(*sa)) < 0) {
+		close(this->fd);
+		this->fd = -1;
+		return -1;
 	}
 
 	return 0;
 }
 
+
+static int last_proxy_port = 0;
+
+/*
+ *	Externally visible function for creating a new proxy LISTENER.
+ *
+ *	For now, don't take ipaddr or port.
+ */
+int proxy_new_listener(void)
+{
+	int port;
+	rad_listen_t *this;
+
+	this = rad_malloc(sizeof(*this));
+
+	memset(this, 0, sizeof(*this));
+
+	this->ipaddr = mainconfig.myip;
+	this->type = RAD_LISTEN_PROXY;
+
+	/*
+	 *	Proxying was not previously defined: die.
+	 */
+	if (last_proxy_port == 0) return -1;
+
+	/*
+	 *	Keep going until we find an unused port.
+	 */
+	for (port = last_proxy_port + 1; port < 64000; port++) {
+		this->port = port;
+		if (listen_bind(this) == 0) {
+			rad_listen_t **last;
+
+			last_proxy_port = port;
+
+			/*
+			 *	Add the new listener to the list of
+			 *	listeners.
+			 */
+			for (last = &mainconfig.listen;
+			     *last != NULL;
+			     last = &((*last)->next)) {
+				/* do nothing */
+			}
+
+			*last = this;
+			return this->fd;
+		}
+	}
+
+	return -1;
+}
+
+
+/*
+ *	Generate a list of listeners.  Takes an input list of
+ *	listeners, too, so we don't close sockets with waiting packets.
+ */
+static int listen_init(const char *filename, rad_listen_t **head)
+{
+	CONF_SECTION	*cs;
+	rad_listen_t	**last;
+	char		buffer[32];
+	rad_listen_t	*this;
+
+	/*
+	 *	Add to the end of the list.
+	 */
+	for (last = head; *last != NULL; last = &((*last)->next)) {
+		/* do nothing */
+	}
+
+    	/*
+	 *	Find the first one (if any).
+	 */
+	for (cs = cf_subsection_find_next(mainconfig.config,
+					  NULL, "listen");
+	     cs != NULL;
+	     cs = cf_subsection_find_next(mainconfig.config,
+					  cs, "listen")) {
+		memset(&listen_inst, 0, sizeof(listen_inst));
+		
+		/*
+		 *	Fix errors for later.
+		 */
+		if (cf_section_parse(cs, &listen_inst, listen_config) < 0) {
+			radlog(L_CONS|L_ERR, "%s[%d]: Error parsing listen section.",
+			       filename, cf_section_lineno(cs));
+			return -1;
+		}
+
+		if (listen_type) {
+			listen_inst.type = lrad_str2int(listen_compare,
+							listen_type, 0);
+		}
+		if (listen_inst.type == RAD_LISTEN_NONE) {
+			radlog(L_CONS|L_ERR, "%s[%d]: Invalid type in listen section.",
+			       filename, cf_section_lineno(cs));
+			return -1;
+		}
+
+		this = rad_malloc(sizeof(*this));
+		memcpy(this, &listen_inst, sizeof(*this));
+		
+		/*
+		 *	And bind it to the port.
+		 */
+		if (listen_bind(this) < 0) {
+			radlog(L_CONS|L_ERR, "%s[%d]: Error binding to port for %s:%d",
+			       filename, cf_section_lineno(cs),
+			       ip_ntoa(buffer, this->ipaddr), this->port);
+			free(this);
+			return -1;
+		}
+
+		*last = this;
+		last = &(this->next);		
+	}
+
+	/*
+	 *	If we're proxying requests, open the proxy FD.
+	 *	Otherwise, don't do anything.
+	 */
+	if (mainconfig.proxy_requests == TRUE) {
+		int		port = -1;
+		rad_listen_t	*auth;
+
+		/*
+		 *	Find the first authentication port,
+		 *	and use it
+		 */
+		for (auth = *head; auth != NULL; auth = auth->next) {
+			if (auth->type == RAD_LISTEN_AUTH) {
+				port = auth->port + 2;
+				break;
+			}
+		}
+
+		/*
+		 *	Not found, pick an accounting port.
+		 */
+		if (port < 0) for (auth = *head; auth != NULL; auth = auth->next) {
+			if (auth->type == RAD_LISTEN_ACCT) {
+				port = auth->port + 1;
+				break;
+			}
+		}
+
+		/*
+		 *	Still no port.  Don't do anything.
+		 */
+		if (port < 0) {
+			return 0;
+		}
+
+		this = rad_malloc(sizeof(*this));
+		memset(this, 0, sizeof(*this));
+		
+		/*
+		 *	Create the proxy socket.
+		 */
+		this->ipaddr = mainconfig.myip;
+		this->type = RAD_LISTEN_PROXY;
+
+		/*
+		 *	Try to find a proxy port (value doesn't matter)
+		 */
+		for (this->port = port;
+		     this->port < 64000;
+		     this->port++) {
+			if (listen_bind(this) == 0) {
+				last_proxy_port = this->port;
+				*last = this;
+				return 0;
+			}
+		}
+
+		radlog(L_ERR|L_CONS, "Failed to open socket for proxying");
+		free(this);
+		return -1;
+	}
+
+	return 0;
+}
+
+
+/*
+ *	Hack the OLD way of listening on a socket.
+ */
+static int old_listen_init(rad_listen_t **head)
+{
+	CONF_PAIR	*cp;
+	rad_listen_t 	*this, **last;
+
+	/*
+	 *	No "bind_address": all listen directives
+	 *	are in the "listen" clauses.
+	 */
+	cp = cf_pair_find(mainconfig.config, "bind_address");
+	if (!cp) return 0;
+	
+	last = head;
+
+	this = rad_malloc(sizeof(*this));
+	memset(this, 0, sizeof(*this));
+
+	/*
+	 *	Create the authentication socket.
+	 */
+       	this->ipaddr = mainconfig.myip;
+	this->type = RAD_LISTEN_AUTH;
+	this->port = auth_port;
+
+	if (listen_bind(this) < 0) {
+		radlog(L_CONS|L_ERR, "There appears to be another RADIUS server running on the authentication port %d", this->port);
+		free(this);
+		return -1;
+	}
+	auth_port = this->port;	/* may have been updated in listen_bind */
+	*last = this;
+	last = &(this->next);
+
+	/*
+	 *  Open Accounting Socket.
+	 *
+	 *  If we haven't already gotten acct_port from /etc/services,
+	 *  then make it auth_port + 1.
+	 */
+	this = rad_malloc(sizeof(*this));
+	memset(this, 0, sizeof(*this));
+
+	/*
+	 *	Create the accounting socket.
+	 *
+	 *	The accounting port is always the authentication port + 1
+	 */
+       	this->ipaddr = mainconfig.myip;
+	this->type = RAD_LISTEN_ACCT;
+	this->port = auth_port + 1;
+
+	if (listen_bind(this) < 0) {
+		radlog(L_CONS|L_ERR, "There appears to be another RADIUS server running on the accounting port %d", this->port);
+		free(this);
+		return -1;
+	}
+	*last = this;
+
+	return 0;
+}
+
+
 #ifndef RADIUS_CONFIG
 #define RADIUS_CONFIG "radiusd.conf"
 #endif
@@ -730,6 +1157,8 @@ int read_mainconfig(int reload)
 	static int old_debug_level = -1;
 	char buffer[1024];
 	CONF_SECTION *cs, *oldcs;
+	rad_listen_t *listener;
+	RADCLIENT *c, *tail;
 
 	if (!reload) {
 		radlog(L_INFO, "Starting - reading configuration files ...");
@@ -779,10 +1208,20 @@ int read_mainconfig(int reload)
 	 *	Add to that, the *new* list of clients.
 	 */
 	snprintf(buffer, sizeof(buffer), "%.200s/%.50s", radius_dir, RADIUS_CONFIG);
-	if (generate_clients(buffer) < 0) {
+	c = generate_clients(buffer, mainconfig.config);
+	if (!c) {
 		return -1;
 	}
 
+	/*
+	 *	The new list of clients takes precedence over the old one.
+	 */
+	for (tail = c; tail->next != NULL; tail = tail->next) {
+	  /* do nothing */
+	}
+	tail->next = mainconfig.clients;
+	mainconfig.clients = c;
+	
 	/* old-style realms file */
 	snprintf(buffer, sizeof(buffer), "%.200s/%.50s", radius_dir, RADIUS_REALMS);
 	DEBUG2("read_config_files:  reading realms");
@@ -819,7 +1258,7 @@ int read_mainconfig(int reload)
 	}
 	librad_debug = debug_flag;
 	old_debug_level = mainconfig.debug_level;
-	
+
 	/*
 	 *  Go update our behaviour, based on the configuration
 	 *  changes.
@@ -854,7 +1293,7 @@ int read_mainconfig(int reload)
 
 		limits.rlim_cur = 0;
 		limits.rlim_max = core_limits.rlim_max;
-		
+
 		if (setrlimit(RLIMIT_CORE, &limits) < 0) {
 			radlog(L_ERR|L_CONS, "Cannot disable core dumps: %s",
 					strerror(errno));
@@ -871,7 +1310,7 @@ int read_mainconfig(int reload)
 		 *	We need root to do mkdir() and chown(), so we
 		 *	do this before giving up root.
 		 */
-		radlogdir_iswritable(mainconfig.uid_name); 
+		radlogdir_iswritable(mainconfig.uid_name);
 	}
 	switch_users();
 
@@ -883,6 +1322,30 @@ int read_mainconfig(int reload)
 		mainconfig.reject_delay = mainconfig.cleanup_delay;
 	}
 
+	/*
+	 *	Initialize the old "bind_address" and "port", first.
+	 */
+	listener = NULL;
+	if (old_listen_init(&listener) < 0) {
+		exit(1);
+	}
+
+	/*
+	 *	Read the list of listeners.
+	 */
+	snprintf(buffer, sizeof(buffer), "%.200s/radiusd.conf", radius_dir);
+	if (listen_init(buffer, &listener) < 0) {
+		exit(1);
+	}
+
+	if (!listener) {
+		radlog(L_ERR|L_CONS, "Server is not configured to listen on any ports.  Exiting.");
+		exit(1);
+	}
+
+	listen_free(mainconfig.listen);
+	mainconfig.listen = listener;
+
 	return 0;
 }
 
