@@ -1,7 +1,7 @@
 /*
  * modcall.c
  *
- * Version:	$Id: modcall.c,v 1.15.2.1 2003/09/28 13:51:25 phampson Exp $
+ * Version:	$Id: modcall.c,v 1.22 2004/04/19 15:54:32 kkalev Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -36,8 +36,11 @@ static modcallable *do_compile_modgroup(int, CONF_SECTION *, const char *,
 
 /* Actions may be a positive integer (the highest one returned in the group
  * will be returned), or the keyword "return", represented here by
- * MOD_ACTION_RETURN, to cause an immediate return. */
+ * MOD_ACTION_RETURN, to cause an immediate return.
+ * There's also the keyword "reject", represented here by MOD_ACTION_REJECT
+ * to cause an immediate reject. */
 #define MOD_ACTION_RETURN (-1)
+#define MOD_ACTION_REJECT (-2)
 
 /* Here are our basic types: modcallable, modgroup, and modsingle. For an
  * explanation of what they are all about, see ../../doc/README.failover */
@@ -129,9 +132,22 @@ static int str2action(const char *s, const char *filename, int lineno)
 {
 	if(!strcasecmp(s, "return"))
 		return MOD_ACTION_RETURN;
-	else if(strspn(s, "0123456789")==strlen(s))
-		return atoi(s);
-	else {
+	else if(!strcasecmp(s, "reject"))
+		return MOD_ACTION_REJECT;
+	else if (strspn(s, "0123456789")==strlen(s)) {
+		int rcode = atoi(s);
+
+		/*
+		 *	Don't allow priority zero, for future use.
+		 */
+		if (rcode == 0) {
+			radlog(L_ERR|L_CONS,
+			       "%s[%d] Invalid action '%s'.\n",
+			       filename, lineno, s);
+			exit(1);
+		}
+		return rcode;
+	} else {
 		radlog(L_ERR|L_CONS,
 				"%s[%d] Unknown action '%s'.\n",
 				filename, lineno, s);
@@ -146,6 +162,8 @@ static const char *action2str(int action)
 	static char buf[32];
 	if(action==MOD_ACTION_RETURN)
 		return "return";
+	if(action==MOD_ACTION_REJECT)
+		return "reject";
 	snprintf(buf, sizeof buf, "%d", action);
 	return buf;
 }
@@ -163,13 +181,13 @@ static const char *comp2str[] = {
 	"post-auth"
 };
 
-#if HAVE_PTHREAD_H
+#ifdef HAVE_PTHREAD_H
 /*
  *	Lock the mutex for the module
  */
 static void safe_lock(module_instance_t *instance)
 {
-	if (instance->mutex) 
+	if (instance->mutex)
 		pthread_mutex_lock(instance->mutex);
 }
 
@@ -178,7 +196,7 @@ static void safe_lock(module_instance_t *instance)
  */
 static void safe_unlock(module_instance_t *instance)
 {
-	if (instance->mutex) 
+	if (instance->mutex)
 		pthread_mutex_unlock(instance->mutex);
 }
 #else
@@ -190,19 +208,19 @@ static void safe_unlock(module_instance_t *instance)
 #endif
 
 static int call_modsingle(int component, modsingle *sp, REQUEST *request,
-		int default_result)
+			  int default_result)
 {
 	int myresult = default_result;
 
 	DEBUG3("  modsingle[%s]: calling %s (%s) for request %d",
-	       comp2str[component], sp->modinst->name, 
+	       comp2str[component], sp->modinst->name,
 	       sp->modinst->entry->name, request->number);
 	safe_lock(sp->modinst);
 	myresult = sp->modinst->entry->module->methods[component](
 			sp->modinst->insthandle, request);
 	safe_unlock(sp->modinst);
 	DEBUG3("  modsingle[%s]: returned from %s (%s) for request %d",
-	       comp2str[component], sp->modinst->name, 
+	       comp2str[component], sp->modinst->name,
 	       sp->modinst->entry->name, request->number);
 
 	return myresult;
@@ -215,6 +233,14 @@ static int call_modgroup(int component, modgroup *g, REQUEST *request,
 	int myresultpref;
 	modcallable *p;
 
+	/*
+	 *	Catch people who have issues.
+	 */
+	if (!g->children) {
+		DEBUG2("  WARNING! Asked to process empty group.  Returning %s.", lrad_int2str(rcode_table, myresult, "??"));
+		return default_result;
+	}
+
 	/* Assign the lowest possible preference to the default return code */
 	myresultpref = 0;
 
@@ -239,6 +265,12 @@ static int call_modgroup(int component, modgroup *g, REQUEST *request,
 			break;
 		}
 
+		/* If "reject" break out of the loop and return reject */
+		if (p->actions[r] == MOD_ACTION_REJECT) {
+			myresult = RLM_MODULE_REJECT;
+			break;
+		}
+
 		/* Otherwise, the action is a number, the preference level of
 		 * this return code. If no higher preference has been seen
 		 * yet, remember this one. */
@@ -298,14 +330,14 @@ int modcall(int component, modcallable *c, REQUEST *request)
 		modgroup *g = mod_callabletogroup(c);
 
 		DEBUG2("modcall: entering group %s for request %d",
-				c->name, request->number);
+		       c->name, request->number);
 
 		myresult = call_modgroup(component, g, request, myresult);
 
 		DEBUG2("modcall: group %s returns %s for request %d",
-		       c->name,
+		        c->name,
 		       lrad_int2str(rcode_table, myresult, "??"),
-			   request->number);
+		       request->number);
 	} else {
 		modsingle *sp = mod_callabletosingle(c);
 
@@ -314,7 +346,7 @@ int modcall(int component, modcallable *c, REQUEST *request)
 		DEBUG2("  modcall[%s]: module \"%s\" returns %s for request %d",
 		       comp2str[component], c->name,
 		       lrad_int2str(rcode_table, myresult, "??"),
-			   request->number);
+		       request->number);
 	}
 
 	return myresult;
@@ -355,7 +387,7 @@ static void dump_tree(int comp, modcallable *c)
 	dump_mc(c, 0);
 }
 #else
-static void dump_tree(int comp, modcallable *c)
+static void dump_tree(int comp UNUSED, modcallable *c UNUSED)
 {
 	return;
 }
