@@ -20,13 +20,13 @@
  *   You should have received a copy of the GNU General Public License
  *   along with this program; if not, write to the Free Software
  *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * 
+ *
  * Copyright 2001 Cistron Internet Services B.V.
  * Copyright 2002 Simon Ekstrand <simon@routemeister.net>
  *
  */
 char radrelay_rcsid[] =
-"$Id: radrelay.c,v 1.13.2.2 2003/08/31 14:01:19 phampson Exp $";
+"$Id: radrelay.c,v 1.22 2004/04/28 21:22:40 kkalev Exp $";
 
 #include "autoconf.h"
 #include "libradius.h"
@@ -35,7 +35,7 @@ char radrelay_rcsid[] =
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <sys/stat.h>
-#if HAVE_NETINET_IN_H
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
 #endif
 
@@ -64,8 +64,6 @@ radlog_dest_t radlog_dest = RADLOG_FILES;
 const char *radius_dir = NULL;
 const char *radacct_dir = NULL;
 const char *radlib_dir = NULL;
-int auth_port = 0;
-int acct_port;
 uint32_t myip = INADDR_ANY;
 int log_stripped_names;
 struct main_config_t mainconfig;
@@ -94,6 +92,7 @@ struct main_config_t mainconfig;
 struct relay_request {
 	int		state;				/* REQ_* state */
 	time_t		retrans;			/* when to retrans */
+	unsigned int	retrans_num;			/* Number of retransmissions */
 	time_t		timestamp;			/* orig recv time */
 	uint32_t	client_ip;			/* Client-IP-Addr */
 	RADIUS_PACKET	*req;				/* Radius request */
@@ -235,6 +234,7 @@ int read_one(FILE *fp, struct relay_request *r_req)
 	if (x == -1)
 		return 0;
 
+redo:
 	s = NULL;
 	fseek(fp, fpos, SEEK_SET);
 	fpos = ftell(fp);
@@ -247,7 +247,11 @@ int read_one(FILE *fp, struct relay_request *r_req)
 		 * being flushed properly. Things should be ok next time
 		 * around.
 		 */
-		if (buf[strlen(buf) - 1] != '\n') {
+		if (!strlen(buf)) {
+			fprintf(stdout, "read_one: ZERO BYTE\n");
+                       fseek(fp, fpos + 1, SEEK_SET);
+                       break;
+               } else if (buf[strlen(buf) - 1] != '\n') {
 			fprintf(stdout, "read_one: BROKEN ATTRIBUTE\n");
 			fseek(fp, fpos + strlen(buf), SEEK_SET);
 			break;
@@ -298,6 +302,28 @@ int read_one(FILE *fp, struct relay_request *r_req)
 		/*
 		 *	w00 - we just completed reading a record in full.
 		 */
+
+		/*
+		 * Check that we have an Acct-Status-Type attribute. If not
+		 * reject the record
+		 */
+		if (pairfind(r_req->req->vps, PW_ACCT_STATUS_TYPE) == NULL){
+			fprintf(stdout, "read_one: No Acct-Status-Type attribute present. Rejecting record.\n");
+			r_req->state = STATE_BUSY1;
+			if (r_req->req->vps != NULL) {
+				pairfree(&r_req->req->vps);
+				r_req->req->vps = NULL;
+			}
+			if (r_req->req->data != NULL) {
+				free (r_req->req->data);
+				r_req->req->data = NULL;
+			}
+			r_req->retrans = 0;
+			r_req->retrans_num = 0;
+			r_req->timestamp = 0;
+			r_req->client_ip = 0;
+			goto redo;
+		}
 		if (r_req->timestamp == 0)
 			r_req->timestamp = time(NULL);
 		if ((vp = pairfind(r_req->req->vps, PW_ACCT_DELAY_TIME)) != NULL) {
@@ -383,6 +409,7 @@ int do_recv(struct relay_misc *r_args)
 			}
 			r->state = 0;
 			r->retrans = 0;
+			r->retrans_num = 0;
 			r->timestamp = 0;
 			r->client_ip = 0;
 			break;
@@ -419,6 +446,7 @@ int do_send(struct relay_request *r, char *secret)
 		}
 		r->state = 0;
 		r->retrans = 0;
+		r->retrans_num = 0;
 		r->timestamp = 0;
 		r->client_ip = 0;
 		return 0;
@@ -437,7 +465,7 @@ int do_send(struct relay_request *r, char *secret)
 	 * Acct-Delay-Time value).
 	 * Otherwise the radius server may consider the
 	 * packet a duplicate and we 'll get caught in a
-	 * loop. 
+	 * loop.
 	 */
 	if (r->retrans > 0){
 		id_map[r->req->id] = 0;
@@ -446,8 +474,12 @@ int do_send(struct relay_request *r, char *secret)
 			free(r->req->data);
 			r->req->data = NULL;
 		}
+		r->retrans_num++;
 	}
-	r->retrans = now + 3;
+	if (r->retrans_num > 20)
+		r->retrans = now + 70;
+	else
+		r->retrans = now + 3 + (3 * r->retrans_num);
 
 	/*
 	 *	Find the Acct-Delay-Time attribute. If it's
@@ -532,6 +564,7 @@ void loop(struct relay_misc *r_args)
 		}
 		slots[i].state = 0;
 		slots[i].retrans = 0;
+		slots[i].retrans_num = 0;
 		slots[i].timestamp = 0;
 		slots[i].client_ip = 0;
 		slots[i].req->sockfd = r_args->sockfd;
@@ -953,7 +986,7 @@ int main(int argc, char **argv)
 		signal(SIGHUP,  SIG_IGN);
 		signal(SIGINT,  SIG_IGN);
 		signal(SIGQUIT, SIG_IGN);
-#if HAVE_SETSID
+#ifdef HAVE_SETSID
 		setsid();
 #endif
 	}
