@@ -1,7 +1,7 @@
 /*
  * radzap.c	Zap a user from the radutmp and radwtmp file.
  *
- * Version:	$Id: radzap.c,v 1.37 2003/03/18 05:50:54 aland Exp $
+ * Version:	$Id: radzap.c,v 1.40 2004/04/06 19:10:59 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -35,7 +35,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 
-#if HAVE_NETINET_IN_H
+#ifdef HAVE_NETINET_IN_H
 #  include <netinet/in.h>
 #endif
 
@@ -50,7 +50,6 @@ const char *radacct_dir = NULL;
 const char *radlib_dir = NULL;
 radlog_dest_t radlog_dest = RADLOG_FILES;
 int debug_flag = 0;
-int auth_port = 0;
 int acct_port;
 int log_stripped_names;
 struct main_config_t mainconfig;
@@ -117,7 +116,7 @@ static void usage(void)
         fprintf(stderr, "  [port]          Terminal Server port to match\n");
         fprintf(stderr, "  [user]          Login account to match\n");
         exit(1);
-}                
+}
 
 
 /*
@@ -141,9 +140,9 @@ int main(int argc, char **argv)
 
         /*  Process the options.  */
         while ((argval = getopt(argc, argv, "d:p:r:")) != EOF) {
-                                
+
                 switch(argval) {
-                        
+
 			case 'd':
 				if (radius_dir) free(radius_dir);
 				radius_dir = strdup(optarg);
@@ -151,7 +150,7 @@ int main(int argc, char **argv)
                         case 'p':
 				acct_port = atoi(optarg);
                                 break;
-                                 
+
                         case 'r':
                                 if ((radiusip = ip_getaddr(optarg)) == INADDR_NONE) {
                                         fprintf(stderr, "%s: %s: radius server unknown\n",
@@ -159,7 +158,7 @@ int main(int argc, char **argv)
                                         exit(1);
 				}
                                 break;
-                               
+
                         default:
                                 usage();
                                 exit(1);
@@ -191,7 +190,7 @@ int main(int argc, char **argv)
 			exit(1);
 		}
 	}
-	if (nas != NULL) 
+	if (nas != NULL)
 		ip = nas->ipaddr;
 
 	/*
@@ -217,7 +216,7 @@ int main(int argc, char **argv)
 
 	printf("%s: zapping termserver %s, port %u",
 		progname, ip_hostname(buf, sizeof(buf), ip), nas_port);
-	if (user != NULL) 
+	if (user != NULL)
 		printf(", user %s", user);
 	printf("\n");
 
@@ -298,9 +297,9 @@ static int do_packet(int allports, uint32_t nasaddr, const struct radutmp *u)
 	req->id = getpid() & 0xFF;
 	req->code = PW_ACCOUNTING_REQUEST;
 	req->dst_port = acct_port;
-	if(req->dst_port == 0) 
+	if(req->dst_port == 0)
 		req->dst_port = getport("radacct");
-	if(req->dst_port == 0) 
+	if(req->dst_port == 0)
 		req->dst_port = PW_ACCT_UDP_PORT;
 	if (radiusip == INADDR_NONE) {
 		req->dst_ipaddr = ip_getaddr("localhost");
@@ -308,7 +307,7 @@ static int do_packet(int allports, uint32_t nasaddr, const struct radutmp *u)
 	else {
 		req->dst_ipaddr = radiusip;
 	}
-	if(!req->dst_ipaddr) 
+	if(!req->dst_ipaddr)
 		req->dst_ipaddr = 0x7f000001;
 	req->vps = NULL;
 	secret = getsecret(req->dst_ipaddr);
