@@ -2,7 +2,7 @@
  * valuepair.c	Valuepair functions that are radiusd-specific
  *		and as such do not belong in the library.
  *
- * Version:	$Id: valuepair.c,v 1.60.2.2 2004/07/14 21:01:02 aland Exp $
+ * Version:	$Id: valuepair.c,v 1.60.2.3 2004/09/09 14:31:06 aland Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  * Copyright 2000  Alan DeKok <aland@ox.org>
  */
 
-static const char rcsid[] = "$Id: valuepair.c,v 1.60.2.2 2004/07/14 21:01:02 aland Exp $";
+static const char rcsid[] = "$Id: valuepair.c,v 1.60.2.3 2004/09/09 14:31:06 aland Exp $";
 
 #include "autoconf.h"
 #include "libradius.h"
@@ -366,13 +366,14 @@ int paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check, VALUE_PAIR **r
 					REG_EXTENDED);
 				compare = regexec(&reg,
 						  (char *)auth_item->strvalue,
-						  16, rxmatch, 0);
+						  REQUEST_MAX_REGEX + 1,
+						  rxmatch, 0);
 				regfree(&reg);
 
 				/*
 				 *	Add %{0}, %{1}, etc.
 				 */
-				for (i = 0; i <= 8; i++) {
+				for (i = 0; i <= REQUEST_MAX_REGEX; i++) {
 					char *p;
 					char buffer[sizeof(check_item->strvalue)];
 
