@@ -1,18 +1,31 @@
 /*
- * rlm_ldap.c LDAP authorization and authentication module.
- * 
- * 
- * This module is based on LDAP patch to Cistron radiusd by James Golovich 
- * <james@wwnet.net>, which in turn was based mostly on a Mysql+Cistron patch 
+ * rlm_ldap.c	LDAP authorization and authentication module.
+ *
+ *   This program is free software; you can redistribute it and/or modify
+ *   it under the terms of the GNU General Public License as published by
+ *   the Free Software Foundation; either version 2 of the License, or
+ *   (at your option) any later version.
+ *
+ *   This program is distributed in the hope that it will be useful,
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *   GNU General Public License for more details.
+ *
+ *   You should have received a copy of the GNU General Public License
+ *   along with this program; if not, write to the Free Software
+ *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ * This module is based on LDAP patch to Cistron radiusd by James Golovich
+ * <james@wwnet.net>, which in turn was based mostly on a Mysql+Cistron patch
  * from <oyarzun@wilmington.net>
- * 
+ *
  * 17 Jan 2000,	Adrian Pavlykevych <pam@polynet.lviv.ua>
  *	- OpenLDAP SDK porting, basic TLS support, LDAP authorization,
- *	  fault tolerance with multiple LDAP server support 
- * 24 May 2000,	Adrian Pavlykevych <pam@polynet.lviv.ua> 
+ *	  fault tolerance with multiple LDAP server support
+ * 24 May 2000,	Adrian Pavlykevych <pam@polynet.lviv.ua>
  *	- Converting to new configuration file format, futher improvements
  *	  in fault tolerance, threaded operation
- * 12 Dec 2000,	Adrian Pavlykevych <pam@polynet.lviv.ua> 
+ * 12 Dec 2000,	Adrian Pavlykevych <pam@polynet.lviv.ua>
  *	- Added preliminary support for multiple instances
  * 	- moved all instance configuration into dynamicly allocated structure
  *	- Removed connection maintenance thread and all attempts for multihreading
@@ -25,9 +38,9 @@
  *  	- Support for generic RADIUS check and reply attribute.
  * Jun 2001, Kostas Kalevras <kkalev@noc.ntua.gr>
  *	- Fix: check and reply attributes from LDAP _replace_ existing ones
- *	- Added "default_profile" directive, which points to radiusProfile 
+ *	- Added "default_profile" directive, which points to radiusProfile
  *	  object, which contains default values for RADIUS users
- *	- Added "profile_attribute" directive, which specifies user object 
+ *	- Added "profile_attribute" directive, which specifies user object
  *	  attribute pointing to radiusProfile object.
  * Nov 2001, Kostas Kalevras <kkalev@noc.ntua.gr>
  *	- Added support for adding the user password to the check. Based on
@@ -117,8 +130,34 @@
  *	- Don't do a double free on the attribute maps. Bug noted by Derrik Pates <dpates@dsdk12.net>
  *	- Apply a patch from Alexander M. Pravking <fduch@antar.bryansk.ru> to do an xlat on the
  *	  retrieved attributes.
+ * Aug 2003, Kostas Kalevras <kkalev@noc.ntua.gr>
+ *	- In case of a bad search filter, print out the corresponding filter
+ * Sep 2003, Kostas Kalevras <kkalev@noc.ntua.gr>
+ *	- Compile even if we don't have pthread's
+ * Oct 2003, Kostas Kalevras <kkalev@noc.ntua.gr>
+ *	- Add a new configuration directive, base_filter which is used for base scope searches
+ *	  (When searching for the default/regular profiles for example)
+ * Nov 2003, Kostas Kalevras <kkalev@noc.ntua.gr>
+ *	- Add a new configuration directive, do_xlat (default: yes). If set we use pairxlatmove
+ *	  on the radius attributes, else we fall back to the plain old pairadd. That way people
+ *	  can fall back on the 0.8.1 behaviour without making changes to their ldap database or
+ *	  gain a little performance by not using pairxlatmove
+ * Dec 2003, Kostas Kalevras <kkalev@noc.ntua.gr>
+ *	- Add a patch from Jon Miner <miner@doit.wisc.edu> to add the ability to configure
+ *	  various LDAP TLS options
+ *	- Only call pairfree if we are using pairxlatmove not for pairadd
+ * Mar 2004, Kostas Kalevras <kkalev@noc.ntua.gr>
+ *	- If we are passed an empty password log a module failure message not an error message
+ * Apr 2004, Kostas Kalveras <kkalev@noc.ntua.gr>
+ *	- Add a patch from Tarun Bhushan <tarun.bhushan@macquarie.com> to add a tls_mode boolean
+ *	  directive so that we can enable TLS connetions even if port is not set to 636
+ *	- Add an error message if ldap_initialize() is not available and we are passed a URL like
+ *	  'server' directive.
+ *	- Add a per instance Ldap-Group attribute (of the form <instance>-Ldap-Group) and register
+ *	  a corresponding ldap_groupcmp function
+ *	- Small change to ldap_get_conn to fix problems on some platforms
  */
-static const char rcsid[] = "$Id: rlm_ldap.c,v 1.108 2003/07/07 19:07:08 aland Exp $";
+static const char rcsid[] = "$Id: rlm_ldap.c,v 1.122.2.1 2004/07/06 19:29:41 aland Exp $";
 
 #include "autoconf.h"
 
@@ -148,6 +187,18 @@ static const char rcsid[] = "$Id: rlm_ldap.c,v 1.108 2003/07/07 19:07:08 aland E
 #include	"modules.h"
 #include	"rad_assert.h"
 
+#ifndef HAVE_PTHREAD_H
+/*
+ *      This is a lot simpler than putting ifdef's around
+ *      every use of the pthread functions.
+ */
+#define pthread_mutex_lock(a)
+#define pthread_mutex_unlock(a)
+#define pthread_mutex_init(a,b)
+#define pthread_mutex_destroy(a)
+#endif
+
+
 #define MAX_FILTER_STR_LEN	1024
 #define TIMELIMIT 5
 
@@ -207,12 +258,14 @@ typedef struct {
 	int		start_tls;
 	int		num_conns;
 	int		do_comp;
+	int		do_xlat;
 	int		default_allow;
 	int		failed_conns;
 	int		is_url;
 	char           *login;
 	char           *password;
 	char           *filter;
+	char           *base_filter;
 	char           *basedn;
 	char           *default_profile;
 	char           *profile_attr;
@@ -229,8 +282,17 @@ typedef struct {
 	LDAP_CONN	*conns;
 	int             ldap_debug; /* Debug flag for LDAP SDK */
 	char		*xlat_name; /* name used to xlat */
+	char		*tls_cacertfile;
+	char		*tls_cacertdir;
+	char		*tls_certfile;
+	char		*tls_keyfile;
+	char		*tls_randfile;
+	char		*tls_require_cert;
 }               ldap_instance;
 
+/* The default setting for TLS Certificate Verification */
+#define TLS_DEFAULT_VERIFY "allow"
+
 static CONF_PARSER module_config[] = {
 	{"server", PW_TYPE_STRING_PTR, offsetof(ldap_instance,server), NULL, "localhost"},
 	{"port", PW_TYPE_INTEGER, offsetof(ldap_instance,port), NULL, "389"},
@@ -241,10 +303,18 @@ static CONF_PARSER module_config[] = {
 	/* allow server unlimited time for search (server-side limit) */
 	{"timelimit", PW_TYPE_INTEGER, offsetof(ldap_instance,timelimit), NULL, "20"},
 	{"identity", PW_TYPE_STRING_PTR, offsetof(ldap_instance,login), NULL, ""},
+	{"tls_mode", PW_TYPE_BOOLEAN, offsetof(ldap_instance,tls_mode), NULL, "no"},
 	{"start_tls", PW_TYPE_BOOLEAN, offsetof(ldap_instance,start_tls), NULL, "no"},
+	{"tls_cacertfile", PW_TYPE_STRING_PTR, offsetof(ldap_instance,tls_cacertfile), NULL, NULL},
+	{"tls_cacertdir", PW_TYPE_STRING_PTR, offsetof(ldap_instance,tls_cacertdir), NULL, NULL},
+	{"tls_certfile", PW_TYPE_STRING_PTR, offsetof(ldap_instance,tls_certfile), NULL, NULL},
+	{"tls_keyfile", PW_TYPE_STRING_PTR, offsetof(ldap_instance,tls_keyfile), NULL, NULL},
+	{"tls_randfile", PW_TYPE_STRING_PTR, offsetof(ldap_instance,tls_randfile), NULL, NULL},
+	{"tls_require_cert", PW_TYPE_STRING_PTR, offsetof(ldap_instance,tls_require_cert), NULL, TLS_DEFAULT_VERIFY},
 	{"password", PW_TYPE_STRING_PTR, offsetof(ldap_instance,password), NULL, ""},
 	{"basedn", PW_TYPE_STRING_PTR, offsetof(ldap_instance,basedn), NULL, "o=notexist"},
 	{"filter", PW_TYPE_STRING_PTR, offsetof(ldap_instance,filter), NULL, "(uid=%u)"},
+	{"base_filter", PW_TYPE_STRING_PTR, offsetof(ldap_instance,base_filter), NULL, "(objectclass=radiusprofile)"},
 	{"default_profile", PW_TYPE_STRING_PTR, offsetof(ldap_instance,default_profile), NULL, NULL},
 	{"profile_attribute", PW_TYPE_STRING_PTR, offsetof(ldap_instance,profile_attr), NULL, NULL},
 	{"password_header", PW_TYPE_STRING_PTR, offsetof(ldap_instance,passwd_hdr), NULL, NULL},
@@ -260,6 +330,7 @@ static CONF_PARSER module_config[] = {
 	{"ldap_connections_number", PW_TYPE_INTEGER, offsetof(ldap_instance,num_conns), NULL, "5"},
 	{"compare_check_items", PW_TYPE_BOOLEAN, offsetof(ldap_instance,do_comp), NULL, "no"},
 	{"access_attr_used_for_allow", PW_TYPE_BOOLEAN, offsetof(ldap_instance,default_allow), NULL, "yes"},
+	{"do_xlat", PW_TYPE_BOOLEAN, offsetof(ldap_instance,do_xlat), NULL, "yes"},
 
 	{NULL, -1, 0, NULL, NULL}
 };
@@ -282,21 +353,22 @@ static inline int ldap_get_conn(LDAP_CONN *conns,LDAP_CONN **ret,void *instance)
 	ldap_instance *inst = instance;
 	register int i = 0;
 
-	for(;i<inst->num_conns;i++){
+	for(i=0;i<inst->num_conns;i++){
+		DEBUG("rlm_ldap: ldap_get_conn: Checking Id: %d",i);
 		if (conns[i].locked == 0 && pthread_mutex_trylock(&(conns[i].mutex)) == 0){
 			*ret = &conns[i];
 			conns[i].locked = 1;
-			DEBUG("ldap_get_conn: Got Id: %d",i);
+			DEBUG("rlm_ldap: ldap_get_conn: Got Id: %d",i);
 			return i;
 		}
 	}
 
 	return -1;
 }
-	
+
 static inline void ldap_release_conn(int i, LDAP_CONN *conns)
 {
-	DEBUG("ldap_release_conn: Release Id: %d",i);
+	DEBUG("rlm_ldap: ldap_release_conn: Release Id: %d",i);
 	conns[i].locked = 0;
 	pthread_mutex_unlock(&(conns[i].mutex));
 }
@@ -309,7 +381,7 @@ static inline void ldap_release_conn(int i, LDAP_CONN *conns)
  *		 to create an instance of the module.
  *
  *************************************************************************/
-static int 
+static int
 ldap_instantiate(CONF_SECTION * conf, void **instance)
 {
 	ldap_instance  *inst;
@@ -319,6 +391,7 @@ ldap_instantiate(CONF_SECTION * conf, void **instance)
 	int check_map_num = 0;
 	int att_map[3] = {0,0,0};
 	TLDAP_RADIUS *pair;
+	ATTR_FLAGS flags;
 	char *xlat_name;
 
 	inst = rad_malloc(sizeof *inst);
@@ -338,17 +411,21 @@ ldap_instantiate(CONF_SECTION * conf, void **instance)
 		return -1;
 	}
 	inst->is_url = 0;
-#ifdef HAVE_LDAP_INITIALIZE
 	if (ldap_is_ldap_url(inst->server)){
+#ifdef HAVE_LDAP_INITIALIZE
 		inst->is_url = 1;
 		inst->port = 0;
-	}
+#else
+		radlog(L_ERR, "rlm_ldap: 'server' directive is in URL form but ldap_initialize() is not available.");
+		free(inst);
+		return -1;
 #endif
- 
+	}
+
 	inst->timeout.tv_usec = 0;
 	inst->net_timeout.tv_usec = 0;
 	/* workaround for servers which support LDAPS but not START TLS */
-	if(inst->port == LDAPS_PORT)
+	if(inst->port == LDAPS_PORT || inst->tls_mode)
 		inst->tls_mode = LDAP_OPT_X_TLS_HARD;
 	else
 		inst->tls_mode = 0;
@@ -357,15 +434,37 @@ ldap_instantiate(CONF_SECTION * conf, void **instance)
 	inst->conns = NULL;
 	inst->failed_conns = 0;
 
+	DEBUG("rlm_ldap: Registering ldap_groupcmp for Ldap-Group");
 	paircompare_register(PW_LDAP_GROUP, PW_USER_NAME, ldap_groupcmp, inst);
-	DEBUG("conns: %p",inst->conns);
 
 	xlat_name = cf_section_name2(conf);
-	if (xlat_name == NULL) {
+	if (xlat_name != NULL){
+		char *group_name;
+		DICT_ATTR *dattr;
+
+		/*
+		 * Allocate room for <instance>-Ldap-Group
+		 */
+		group_name = malloc((strlen(xlat_name) + 1 + 11) * sizeof(char));
+		rad_assert(group_name != NULL);
+		sprintf(group_name,"%s-Ldap-Group",xlat_name);
+		DEBUG("rlm_ldap: Creating new attribute %s",group_name);
+		dict_addattr(group_name, 0, PW_TYPE_STRING, -1, flags);
+		dattr = dict_attrbyname(group_name);
+		if (dattr == NULL){
+			radlog(L_ERR, "rlm_ldap: Failed to create attribute %s",group_name);
+			free(inst);
+			return -1;
+		}
+		DEBUG("rlm_ldap: Registering ldap_groupcmp for %s",group_name);
+		paircompare_register(dattr->attr, PW_USER_NAME, ldap_groupcmp, inst);
+	}
+	else {
 		xlat_name = cf_section_name1(conf);
 		rad_assert(xlat_name != NULL); /* or all hell breaks loose */
 	}
 	inst->xlat_name = strdup(xlat_name);
+	DEBUG("rlm_ldap: Registering ldap_xlat with xlat_name %s",xlat_name);
 	xlat_register(xlat_name,ldap_xlat,inst);
 
 	if (inst->num_conns <= 0){
@@ -385,14 +484,21 @@ ldap_instantiate(CONF_SECTION * conf, void **instance)
 		inst->conns[i].failed_conns = 0;
 		inst->conns[i].ld = NULL;
 		pthread_mutex_init(&inst->conns[i].mutex, NULL);
-	}	
+	}
 
 	if (read_mappings(inst) != 0) {
 		radlog(L_ERR, "rlm_ldap: Reading dictionary mappings from file %s failed",
 		       inst->dictionary_mapping);
-		radlog(L_ERR, "rlm_ldap: Proceeding with no mappings");
+		free(inst);
+		return -1;
 	}
-	
+	if (inst->check_item_map == NULL && inst->reply_item_map == NULL){
+		radlog(L_ERR, "rlm_ldap: dictionary mappings file %s did not contain any mappings",
+			inst->dictionary_mapping);
+		free(inst);
+		return -1;
+	}
+
 	pair = inst->check_item_map;
 	while(pair != NULL){
 		atts_num++;
@@ -418,6 +524,8 @@ ldap_instantiate(CONF_SECTION * conf, void **instance)
 		return -1;
 	}
 	pair = inst->check_item_map;
+	if (pair == NULL)
+		pair = inst->reply_item_map;
 	for(i=0;i<atts_num;i++){
 		if (i <= check_map_num ){
 			inst->atts[i] = pair->attr;
@@ -445,7 +553,7 @@ ldap_instantiate(CONF_SECTION * conf, void **instance)
 			}
 		}
 	}
-	inst->atts[atts_num] = NULL;		
+	inst->atts[atts_num] = NULL;
 
 	DEBUG("conns: %p",inst->conns);
 
@@ -497,14 +605,14 @@ read_mappings(ldap_instance* inst)
 		/* strip comments */
 		ptr = strchr(buf, '#');
 		if (ptr) *ptr = 0;
-		
+
 		/* empty line */
 		if (buf[0] == 0) continue;
-		
-		/* extract tokens from the string */		
+
+		/* extract tokens from the string */
 		token_count = sscanf(buf, "%s %s %s", itemType, radiusAttribute, ldapAttribute);
 
-		if (token_count <= 0) /* no tokens */			
+		if (token_count <= 0) /* no tokens */
 			continue;
 
 		if (token_count != 3) {
@@ -528,7 +636,7 @@ read_mappings(ldap_instance* inst)
 			fclose(mapfile);
 			return -1;
 		}
-			
+
 		/* push node to correct list */
 		if (strcasecmp(itemType, "checkItem") == 0) {
 			pair->next = inst->check_item_map;
@@ -537,7 +645,7 @@ read_mappings(ldap_instance* inst)
 			pair->next = inst->reply_item_map;
 			inst->reply_item_map = pair;
 		} else {
-			radlog(L_ERR, "rlm_ldap: file %s: skipping line %i: unknown itemType %s", 
+			radlog(L_ERR, "rlm_ldap: file %s: skipping line %i: unknown itemType %s",
 			       filename, linenumber, itemType);
 			free(pair->attr);
 			free(pair->radius_attr);
@@ -548,14 +656,14 @@ read_mappings(ldap_instance* inst)
 		DEBUG("rlm_ldap: LDAP %s mapped to RADIUS %s",
 		      pair->attr, pair->radius_attr);
 	}
-	
+
 	fclose(mapfile);
 
 	return 0; /* success */
 }
 
-static int 
-perform_search(void *instance, LDAP_CONN *conn, char *search_basedn, int scope, char *filter, 
+static int
+perform_search(void *instance, LDAP_CONN *conn, char *search_basedn, int scope, char *filter,
 		char **attrs, LDAPMessage ** result)
 {
 	int             res = RLM_MODULE_OK;
@@ -605,7 +713,7 @@ retry:
 				radlog(L_INFO, "rlm_ldap: Attempting reconnect");
 				search_retry = 1;
 				conn->bound = 0;
-				ldap_msgfree(*result);	
+				ldap_msgfree(*result);
 				goto retry;
 			}
 		}
@@ -619,26 +727,30 @@ retry:
 		radlog(L_ERR, "rlm_ldap: ldap_search() failed: Timed out while waiting for server to respond. Please increase the timeout.");
 		ldap_msgfree(*result);
 		return RLM_MODULE_FAIL;
+	case LDAP_FILTER_ERROR:
+		radlog(L_ERR, "rlm_ldap: ldap_search() failed: Bad search filter: %s",filter);
+		ldap_msgfree(*result);
+		return RLM_MODULE_FAIL;
 	case LDAP_TIMELIMIT_EXCEEDED:
 	case LDAP_BUSY:
 	case LDAP_UNAVAILABLE:
 		/* We don't need to reconnect in these cases so we don't set conn->bound */
 		ldap_get_option(conn->ld, LDAP_OPT_ERROR_NUMBER, &ldap_errno);
 		radlog(L_ERR, "rlm_ldap: ldap_search() failed: %s", ldap_err2string(ldap_errno));
-		ldap_msgfree(*result);	
+		ldap_msgfree(*result);
 		return (RLM_MODULE_FAIL);
 	default:
 		ldap_get_option(conn->ld, LDAP_OPT_ERROR_NUMBER, &ldap_errno);
 		radlog(L_ERR, "rlm_ldap: ldap_search() failed: %s", ldap_err2string(ldap_errno));
 		conn->bound = 0;
-		ldap_msgfree(*result);	
+		ldap_msgfree(*result);
 		return (RLM_MODULE_FAIL);
 	}
 
 	if ((ldap_count_entries(conn->ld, *result)) != 1) {
 		DEBUG("rlm_ldap: object not found or got ambiguous search result");
 		res = RLM_MODULE_NOTFOUND;
-		ldap_msgfree(*result);	
+		ldap_msgfree(*result);
 	}
 	return res;
 }
@@ -650,7 +762,7 @@ retry:
 static int ldap_escape_func(char *out, int outlen, const char *in)
 {
 	int len = 0;
-	
+
 	while (in[0]) {
 		/*
 		 *  Only one byte left.
@@ -658,13 +770,13 @@ static int ldap_escape_func(char *out, int outlen, const char *in)
 		if (outlen <= 1) {
 			break;
 		}
-		
+
 		if (strchr("*", *in)) {
 			in++;
 			outlen--;
 			continue;
 		}
-		
+
 		/*
 		 *	Else it's a nice character.
 		 */
@@ -768,14 +880,14 @@ static int ldap_groupcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALU
 		/* This looks like a DN */
 		snprintf(filter,sizeof(filter), "%s",gr_filter);
 		snprintf(basedn,sizeof(basedn), "%s",(char *)check->strvalue);
-	} else 
+	} else
 		snprintf(filter,sizeof(filter), "(&(%s=%s)%s)",inst->groupname_attr,(char *)check->strvalue,gr_filter);
 
 	if ((conn_id = ldap_get_conn(inst->conns,&conn,inst)) == -1){
 		radlog(L_ERR, "rlm_ldap: All ldap connections are in use");
 		return 1;
 	}
-	
+
 	if ((res = perform_search(inst, conn, basedn, LDAP_SCOPE_SUBTREE,
 				filter, attrs, &result)) == RLM_MODULE_OK){
 		DEBUG("rlm_ldap::ldap_groupcmp: User found in group %s",
@@ -784,7 +896,7 @@ static int ldap_groupcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALU
 		ldap_release_conn(conn_id,inst->conns);
         	return 0;
 	}
-	
+
 	ldap_release_conn(conn_id,inst->conns);
 
 	if (res != RLM_MODULE_NOTFOUND ) {
@@ -793,7 +905,7 @@ static int ldap_groupcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALU
 	}
 
 	if (inst->groupmemb_attr == NULL){
-		/* search returned NOTFOUND and searching for membership 
+		/* search returned NOTFOUND and searching for membership
 		 * using user object attributes is not specified in config
 		 * file
 		 */
@@ -811,7 +923,7 @@ static int ldap_groupcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALU
 		DEBUG("rlm_ldap::ldap_groupcmp: Search returned error");
 		ldap_release_conn(conn_id, inst->conns);
 		return 1;
-	}		
+	}
 
 	if ((msg = ldap_first_entry(conn->ld, result)) == NULL) {
 		DEBUG("rlm_ldap::ldap_groupcmp: ldap_first_entry() failed");
@@ -823,7 +935,7 @@ static int ldap_groupcmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALU
 		unsigned int i = 0;
 		char found = 0;
 		for (;i < ldap_count_values(vals);i++){
-			if (strchr(vals[i],',') != NULL){ 
+			if (strchr(vals[i],',') != NULL){
 				/* This looks like a DN */
 				LDAPMessage *gr_result = NULL;
 				snprintf(filter,sizeof(filter), "(%s=%s)",
@@ -979,7 +1091,7 @@ static int ldap_xlat(void *instance, REQUEST *request, char *fmt, char *out, int
  *      Purpose: Check if user is authorized for remote access
  *
  ******************************************************************************/
-static int 
+static int
 ldap_authorize(void *instance, REQUEST * request)
 {
 	LDAPMessage	*result = NULL;
@@ -1110,7 +1222,7 @@ ldap_authorize(void *instance, REQUEST * request)
 	}
 
 	/*
-	 * Check for the default profile entry. If it exists then add the 
+	 * Check for the default profile entry. If it exists then add the
 	 * attributes it contains in the check and reply pairs
 	 */
 
@@ -1118,33 +1230,39 @@ ldap_authorize(void *instance, REQUEST * request)
 	if (inst->default_profile || user_profile){
 		char *profile = inst->default_profile;
 
-		strNcpy(filter,"(objectclass=radiusprofile)",sizeof(filter));
+		strNcpy(filter,inst->base_filter,sizeof(filter));
 		if (user_profile)
 			profile = user_profile->strvalue;
 		if (profile && strlen(profile)){
 			if ((res = perform_search(instance, conn,
-				profile, LDAP_SCOPE_BASE, 
+				profile, LDAP_SCOPE_BASE,
 				filter, inst->atts, &def_result)) == RLM_MODULE_OK){
 				if ((def_msg = ldap_first_entry(conn->ld,def_result))){
 					if ((check_tmp = ldap_pairget(conn->ld,def_msg,inst->check_item_map,check_pairs,1))) {
-						/*pairadd(check_pairs,check_tmp);*/
-						pairxlatmove(request, check_pairs, &check_tmp);
-						pairfree(&check_tmp);
+						if (inst->do_xlat){
+							pairxlatmove(request, check_pairs, &check_tmp);
+							pairfree(&check_tmp);
+						}
+						else
+							pairadd(check_pairs,check_tmp);
 					}
 					if ((reply_tmp = ldap_pairget(conn->ld,def_msg,inst->reply_item_map,reply_pairs,0))) {
-						/*pairadd(reply_pairs,reply_tmp);*/
-						pairxlatmove(request, reply_pairs, &reply_tmp);
-						pairfree(&reply_tmp);
+						if (inst->do_xlat){
+							pairxlatmove(request, reply_pairs, &reply_tmp);
+							pairfree(&reply_tmp);
+						}
+						else
+							pairadd(reply_pairs,reply_tmp);
 					}
 				}
 				ldap_msgfree(def_result);
-			} else 
+			} else
 				DEBUG("rlm_ldap: default_profile/user-profile search failed");
 		}
 	}
 
 	/*
-	 * Check for the profile attribute. If it exists, we assume that it 
+	 * Check for the profile attribute. If it exists, we assume that it
 	 * contains the DN of an entry containg a profile for the user. That
 	 * way we can have different general profiles for various user groups
 	 * (students,faculty,staff etc)
@@ -1153,25 +1271,31 @@ ldap_authorize(void *instance, REQUEST * request)
 	if (inst->profile_attr){
 		if ((vals = ldap_get_values(conn->ld, msg, inst->profile_attr)) != NULL) {
 			unsigned int i=0;
-			strNcpy(filter,"(objectclass=radiusprofile)",sizeof(filter));
+			strNcpy(filter,inst->base_filter,sizeof(filter));
 			while(vals[i] != NULL && strlen(vals[i])){
 				if ((res = perform_search(instance, conn,
-					vals[i], LDAP_SCOPE_BASE, 
+					vals[i], LDAP_SCOPE_BASE,
 					filter, inst->atts, &def_attr_result)) == RLM_MODULE_OK){
 					if ((def_attr_msg = ldap_first_entry(conn->ld,def_attr_result))){
 						if ((check_tmp = ldap_pairget(conn->ld,def_attr_msg,inst->check_item_map,check_pairs,1))) {
-							/*pairadd(check_pairs,check_tmp);*/
-							pairxlatmove(request, check_pairs, &check_tmp);
-							pairfree(&check_tmp);
+							if (inst->do_xlat){
+								pairxlatmove(request, check_pairs, &check_tmp);
+								pairfree(&check_tmp);
+							}
+							else
+								pairadd(check_pairs,check_tmp);
 						}
 						if ((reply_tmp = ldap_pairget(conn->ld,def_attr_msg,inst->reply_item_map,reply_pairs,0))) {
-							/*pairadd(reply_pairs,reply_tmp);*/
-							pairxlatmove(request, reply_pairs, &reply_tmp);
-							pairfree(&reply_tmp);
+							if (inst->do_xlat){
+								pairxlatmove(request, reply_pairs, &reply_tmp);
+								pairfree(&reply_tmp);
+							}
+							else
+								pairadd(reply_pairs,reply_tmp);
 						}
 					}
 					ldap_msgfree(def_attr_result);
-				} else 
+				} else
 					DEBUG("rlm_ldap: profile_attribute search failed");
 				i++;
 			}
@@ -1226,9 +1350,12 @@ ldap_authorize(void *instance, REQUEST * request)
 	DEBUG("rlm_ldap: looking for check items in directory...");
 
 	if ((check_tmp = ldap_pairget(conn->ld, msg, inst->check_item_map,check_pairs,1)) != NULL) {
-		/*pairadd(check_pairs,check_tmp);*/
-		pairxlatmove(request, check_pairs, &check_tmp);
-		pairfree(&check_tmp);
+		if (inst->do_xlat){
+			pairxlatmove(request, check_pairs, &check_tmp);
+			pairfree(&check_tmp);
+		}
+		else
+			pairadd(check_pairs,check_tmp);
 	}
 
 
@@ -1236,9 +1363,12 @@ ldap_authorize(void *instance, REQUEST * request)
 
 
 	if ((reply_tmp = ldap_pairget(conn->ld, msg, inst->reply_item_map,reply_pairs,0)) != NULL) {
-		/*pairadd(reply_pairs,reply_tmp);*/
-		pairxlatmove(request, reply_pairs, &reply_tmp);
-		pairfree(&reply_tmp);
+		if (inst->do_xlat){
+			pairxlatmove(request, reply_pairs, &reply_tmp);
+			pairfree(&reply_tmp);
+		}
+		else
+			pairadd(reply_pairs,reply_tmp);
 	}
 
        if (inst->do_comp && paircmp(request,request->packet->vps,*check_pairs,reply_pairs) != 0){
@@ -1275,7 +1405,7 @@ ldap_authorize(void *instance, REQUEST * request)
  *	Purpose: Check the user's password against ldap database
  *
  *****************************************************************************/
-static int 
+static int
 ldap_authenticate(void *instance, REQUEST * request)
 {
 	LDAP           *ld_user;
@@ -1314,7 +1444,9 @@ ldap_authenticate(void *instance, REQUEST * request)
 	}
 
 	if (request->password->length == 0) {
-		radlog(L_ERR, "rlm_ldap: empty password supplied");
+		snprintf(module_fmsg,sizeof(module_fmsg),"rlm_ldap: empty password supplied");
+		module_fmsg_vp = pairmake("Module-Failure-Message", module_fmsg, T_OP_EQ);
+		pairadd(&request->packet->vps, module_fmsg_vp);
 		return RLM_MODULE_INVALID;
 	}
 
@@ -1330,13 +1462,13 @@ ldap_authenticate(void *instance, REQUEST * request)
 	}
 
 
-	DEBUG("rlm_ldap: login attempt by \"%s\" with password \"%s\"", 
+	DEBUG("rlm_ldap: login attempt by \"%s\" with password \"%s\"",
 	       request->username->strvalue, request->password->strvalue);
 
 	while((vp_user_dn = pairfind(request->packet->vps, PW_LDAP_USERDN)) == NULL) {
 		if (!radius_xlat(filter, sizeof(filter), inst->filter,
 				request, NULL)) {
-			radlog (L_ERR, "rlm_ldap: unable to create filter.\n"); 
+			radlog (L_ERR, "rlm_ldap: unable to create filter.\n");
 			return RLM_MODULE_INVALID;
 		}
 
@@ -1453,6 +1585,82 @@ ldap_connect(void *instance, const char *dn, const char *password, int auth, int
  			radlog(L_ERR, "rlm_ldap: could not set LDAP_OPT_X_TLS option %s", ldap_err2string(ldap_errno));
 		}
 	}
+
+	if(inst->tls_cacertfile != NULL) {
+		DEBUG("rlm_ldap: setting TLS CACert File to %s", inst->tls_cacertfile);
+
+		if ( ldap_set_option( NULL, LDAP_OPT_X_TLS_CACERTFILE,
+							  (void *) inst->tls_cacertfile )
+			 != LDAP_OPT_SUCCESS) {
+			radlog(L_ERR, "rlm_ldap: could not set "
+				   "LDAP_OPT_X_TLS_CACERTFILE option to %s", inst->tls_cacertfile);
+		}
+	}
+
+	if(inst->tls_cacertdir != NULL) {
+		DEBUG("rlm_ldap: setting TLS CACert File to %s", inst->tls_cacertdir);
+
+		if ( ldap_set_option( NULL, LDAP_OPT_X_TLS_CACERTDIR,
+							  (void *) inst->tls_cacertdir )
+			 != LDAP_OPT_SUCCESS) {
+			radlog(L_ERR, "rlm_ldap: could not set "
+				   "LDAP_OPT_X_TLS_CACERTDIR option to %s", inst->tls_cacertdir);
+		}
+	}
+
+	if( strcmp( TLS_DEFAULT_VERIFY, inst->tls_require_cert ) != 0 ) {
+		DEBUG("rlm_ldap: setting TLS Require Cert to %s",
+			  inst->tls_require_cert);
+	}
+
+
+#ifdef HAVE_INT_TLS_CONFIG
+
+	if ( ldap_int_tls_config( NULL, LDAP_OPT_X_TLS_REQUIRE_CERT,
+							  (inst->tls_require_cert) )
+		 != LDAP_OPT_SUCCESS) {
+		radlog(L_ERR, "rlm_ldap: could not set "
+			   "LDAP_OPT_X_TLS_REQUIRE_CERT option to %s",
+			   inst->tls_require_cert);
+	}
+
+#endif
+
+	if(inst->tls_certfile != NULL) {
+		DEBUG("rlm_ldap: setting TLS Cert File to %s", inst->tls_certfile);
+
+		if ( ldap_set_option( NULL, LDAP_OPT_X_TLS_CERTFILE,
+							  (void *) inst->tls_certfile )
+			 != LDAP_OPT_SUCCESS) {
+			radlog(L_ERR, "rlm_ldap: could not set "
+				   "LDAP_OPT_X_TLS_CERTFILE option to %s",
+				   inst->tls_certfile);
+		}
+	}
+
+	if(inst->tls_keyfile != NULL) {
+		DEBUG("rlm_ldap: setting TLS Key File to %s", inst->tls_keyfile);
+
+		if ( ldap_set_option( NULL, LDAP_OPT_X_TLS_KEYFILE,
+							  (void *) inst->tls_keyfile )
+			 != LDAP_OPT_SUCCESS) {
+			radlog(L_ERR, "rlm_ldap: could not set "
+				   "LDAP_OPT_X_TLS_KEYFILE option to %s",
+				   inst->tls_keyfile);
+		}
+	}
+
+	if(inst->tls_randfile != NULL) {
+		DEBUG("rlm_ldap: setting TLS Key File to %s", inst->tls_randfile);
+
+		if ( ldap_set_option( NULL, LDAP_OPT_X_TLS_RANDOM_FILE,
+							  (void *) inst->tls_randfile )
+			 != LDAP_OPT_SUCCESS) {
+			radlog(L_ERR, "rlm_ldap: could not set "
+				   "LDAP_OPT_X_TLS_RANDOM_FILE option to %s",
+				   inst->tls_randfile);
+		}
+	}
 	if (inst->start_tls) {
 		DEBUG("rlm_ldap: starting TLS");
 		rc = ldap_start_tls_s(ld, NULL, NULL);
@@ -1508,18 +1716,21 @@ ldap_connect(void *instance, const char *dn, const char *password, int auth, int
 	ldap_errno = ldap_result2error(ld, res, 1);
 	switch (ldap_errno) {
 	case LDAP_SUCCESS:
+		DEBUG("rlm_ldap: Bind was successful");
 		*result = RLM_MODULE_OK;
 		break;
 
 	case LDAP_INVALID_CREDENTIALS:
-		if (auth) 
+		if (auth){
+			DEBUG("rlm_ldap: Bind failed with invalid credentials");
 			*result = RLM_MODULE_REJECT;
+		}
 		else {
-			radlog(L_ERR, "rlm_ldap: LDAP login failed: check login, password settings in ldap section of radiusd.conf");
+			radlog(L_ERR, "rlm_ldap: LDAP login failed: check identity, password settings in ldap section of radiusd.conf");
 			*result = RLM_MODULE_FAIL;
-		}		
+		}
 		break;
-		
+
 	default:
 		if (inst->is_url)
 			radlog(L_ERR,"rlm_ldap: %s bind to %s failed %s",
@@ -1543,7 +1754,7 @@ ldap_connect(void *instance, const char *dn, const char *password, int auth, int
  *	Detach from the LDAP server and cleanup internal state.
  *
  *****************************************************************************/
-static int 
+static int
 ldap_detach(void *instance)
 {
 	ldap_instance  *inst = instance;
@@ -1561,6 +1772,8 @@ ldap_detach(void *instance)
 		free(inst->dictionary_mapping);
 	if (inst->filter)
 		free((char *) inst->filter);
+	if (inst->base_filter)
+		free((char *) inst->base_filter);
 	if (inst->passwd_hdr)
 		free((char *) inst->passwd_hdr);
 	if (inst->passwd_attr)
@@ -1588,7 +1801,7 @@ ldap_detach(void *instance)
 	}
 
 	pair = inst->check_item_map;
-	
+
 	while (pair != NULL) {
 		nextpair = pair->next;
 		free(pair->attr);
@@ -1620,7 +1833,7 @@ ldap_detach(void *instance)
 }
 
 #ifdef FIELDCPY
-static void 
+static void
 fieldcpy(char *string, char **uptr)
 {
 	char           *ptr;
@@ -1693,14 +1906,14 @@ ldap_pairget(LDAP * ld, LDAPMessage * entry,
 				if (is_generic_attribute) {
 					/* this is a generic attribute */
 					LRAD_TOKEN dummy; /* makes pairread happy */
-					
+
 					/* not sure if using pairread here is ok ... */
 					if ( (newpair = pairread(&ptr, &dummy)) != NULL) {
-						DEBUG("rlm_ldap: extracted attribute %s from generic item %s", 
+						DEBUG("rlm_ldap: extracted attribute %s from generic item %s",
 						      newpair->name, vals[vals_idx]);
 						pairadd(&pairlist, newpair);
 					} else {
-						radlog(L_ERR, "rlm_ldap: parsing %s failed: %s", 
+						radlog(L_ERR, "rlm_ldap: parsing %s failed: %s",
 						       element->attr, vals[vals_idx]);
 					}
 				} else {
