@@ -3,7 +3,7 @@
  *		Contains the functions for the "huntgroups" and "hints"
  *		files.
  *
- * Version:     $Id: rlm_preprocess.c,v 1.45 2003/07/07 19:17:31 aland Exp $
+ * Version:     $Id: rlm_preprocess.c,v 1.52 2004/05/15 15:30:13 mgriego Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  * Copyright 2000  Alan DeKok <aland@ox.org>
  */
 
-static const char rcsid[] = "$Id: rlm_preprocess.c,v 1.45 2003/07/07 19:17:31 aland Exp $";
+static const char rcsid[] = "$Id: rlm_preprocess.c,v 1.52 2004/05/15 15:30:13 mgriego Exp $";
 
 #include	"autoconf.h"
 #include	"libradius.h"
@@ -116,10 +116,6 @@ static void cisco_vsa_hack(VALUE_PAIR *vp)
 
 	for ( ; vp != NULL; vp = vp->next) {
 		vendorcode = (vp->attribute >> 16); /* HACK! */
-		if (vendorcode == 0) continue;	/* ignore non-VSA attributes */
-
-		if (vendorcode == 0) continue; /* ignore unknown VSA's */
-
 		if (vendorcode != 9) continue; /* not a Cisco VSA, continue */
 
 		if (vp->type != PW_TYPE_STRING) continue;
@@ -134,6 +130,7 @@ static void cisco_vsa_hack(VALUE_PAIR *vp)
 		 *	Cisco-AVPair's get packed as:
 		 *
 		 *	Cisco-AVPair = "h323-foo-bar = baz"
+		 *	Cisco-AVPair = "h323-foo-bar=baz"
 		 *
 		 *	which makes sense only if you're a lunatic.
 		 *	This code looks for the attribute named inside
@@ -145,12 +142,12 @@ static void cisco_vsa_hack(VALUE_PAIR *vp)
 			DICT_ATTR	*dattr;
 
 			p = vp->strvalue;
-			getword(&p, newattr, sizeof(newattr));
+			gettoken(&p, newattr, sizeof(newattr));
 
 			if (((dattr = dict_attrbyname(newattr)) != NULL) &&
 			    (dattr->type == PW_TYPE_STRING)) {
 				VALUE_PAIR *newvp;
-				
+
 				/*
 				 *  Make a new attribute.
 				 */
@@ -188,7 +185,7 @@ static void rad_mangle(rlm_preprocess_t *data, REQUEST *request)
 	 */
 	request_pairs = request->packet->vps;
 	namepair = pairfind(request_pairs, PW_USER_NAME);
-	if ((namepair == NULL) || 
+	if ((namepair == NULL) ||
 	    (namepair->length <= 0)) {
 	  return;
 	}
@@ -250,7 +247,7 @@ static void rad_mangle(rlm_preprocess_t *data, REQUEST *request)
  *	huntgroup, which normally only contains username or group.
  *	At least one of the "reply" items has to match.
  */
-static int hunt_paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
+static int hunt_paircmp(REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check)
 {
 	VALUE_PAIR	*check_item = check;
 	VALUE_PAIR	*tmp;
@@ -263,7 +260,7 @@ static int hunt_paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 		tmp = check_item->next;
 		check_item->next = NULL;
 
-		result = paircmp(NULL, request, check_item, NULL);
+		result = paircmp(req, request, check_item, NULL);
 
 		check_item->next = tmp;
 		check_item = check_item->next;
@@ -273,120 +270,16 @@ static int hunt_paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 }
 
 
-/*
- *	Compare prefix/suffix
- */
-static int presufcmp(VALUE_PAIR *check, char *name, char *rest)
-{
-	int		len, namelen;
-	int		ret = -1;
-
-#if 0 /* DEBUG */
-	printf("Comparing %s and %s, check->attr is %d\n",
-		name, check->strvalue, check->attribute);
-#endif
-
-	len = strlen((char *)check->strvalue);
-	switch (check->attribute) {
-		case PW_PREFIX:
-			ret = strncmp(name, (char *)check->strvalue, len);
-			if (ret == 0 && rest)
-				strcpy(rest, name + len);
-			break;
-		case PW_SUFFIX:
-			namelen = strlen(name);
-			if (namelen < len)
-				break;
-			ret = strcmp(name + namelen - len,
-				     (char *)check->strvalue);
-			if (ret == 0 && rest) {
-				strncpy(rest, name, namelen - len);
-				rest[namelen - len] = 0;
-			}
-			break;
-	}
-
-	return ret;
-}
-
-/*
- *	Match a username with a wildcard expression.
- *	Is very limited for now.
- */
-static int matches(char *name, PAIR_LIST *pl, char *matchpart)
-{
-	int len, wlen;
-	int ret = 0;
-	char *wild = pl->name;
-	VALUE_PAIR *tmp;
-
-	/*
-	 *	We now support both:
-	 *
-	 *		DEFAULT	Prefix = "P"
-	 *
-	 *	and
-	 *		P*
-	 */
-	if ((tmp = pairfind(pl->check, PW_PREFIX)) != NULL ||
-	    (tmp = pairfind(pl->check, PW_SUFFIX)) != NULL) {
-
-		if (strncmp(pl->name, "DEFAULT", 7) == 0 ||
-		    strcmp(pl->name, name) == 0)
-			return !presufcmp(tmp, name, matchpart);
-	}
-
-	/*
-	 *	Shortcut if there's no '*' in pl->name.
-	 */
-	if (strchr(pl->name, '*') == NULL &&
-	    (strncmp(pl->name, "DEFAULT", 7) == 0 ||
-	     strcmp(pl->name, name) == 0)) {
-		strcpy(matchpart, name);
-		return 1;
-	}
-
-	/*
-	 *	Normally, we should return 0 here, but we
-	 *	support the old * stuff.
-	 */
-	len = strlen(name);
-	wlen = strlen(wild);
-
-	if (len == 0 || wlen == 0) return 0;
-
-	if (wild[0] == '*') {
-		wild++;
-		wlen--;
-		if (wlen <= len && strcmp(name + (len - wlen), wild) == 0) {
-			strcpy(matchpart, name);
-			matchpart[len - wlen] = 0;
-			ret = 1;
-		}
-	} else if (wild[wlen - 1] == '*') {
-		if (wlen <= len && strncmp(name, wild, wlen - 1) == 0) {
-			strcpy(matchpart, name + wlen - 1);
-			ret = 1;
-		}
-	}
-
-	return ret;
-}
-
-
 /*
  *	Add hints to the info sent by the terminal server
- *	based on the pattern of the username.
+ *	based on the pattern of the username, and other attributes.
  */
 static int hints_setup(PAIR_LIST *hints, REQUEST *request)
 {
-	char		newname[MAX_STRING_LEN];
 	char		*name;
 	VALUE_PAIR	*add;
-	VALUE_PAIR	*last;
 	VALUE_PAIR	*tmp;
 	PAIR_LIST	*i;
-	int		do_strip;
 	VALUE_PAIR *request_pairs;
 
 	request_pairs = request->packet->vps;
@@ -394,8 +287,8 @@ static int hints_setup(PAIR_LIST *hints, REQUEST *request)
 	if (hints == NULL || request_pairs == NULL)
 		return RLM_MODULE_NOOP;
 
-	/* 
-	 *	Check for valid input, zero length names not permitted 
+	/*
+	 *	Check for valid input, zero length names not permitted
 	 */
 	if ((tmp = pairfind(request_pairs, PW_USER_NAME)) == NULL)
 		name = NULL;
@@ -409,7 +302,10 @@ static int hints_setup(PAIR_LIST *hints, REQUEST *request)
 		return RLM_MODULE_NOOP;
 
 	for (i = hints; i; i = i->next) {
-		if (matches(name, i, newname)) {
+		/*
+		 *	Use "paircmp", which is a little more general...
+		 */
+		if (paircmp(request, request_pairs, i->check, NULL) == 0) {
 			DEBUG2("  hints: Matched %s at %d",
 			       i->name, i->lineno);
 			break;
@@ -420,95 +316,23 @@ static int hints_setup(PAIR_LIST *hints, REQUEST *request)
 
 	add = paircopy(i->reply);
 
-#if 0 /* DEBUG */
-	printf("In hints_setup, newname is %s\n", newname);
-#endif
-
-	/*
-	 *	See if we need to adjust the name.
-	 */
-	do_strip = 1;
-	if ((tmp = pairfind(i->reply, PW_STRIP_USER_NAME)) != NULL
-	     && tmp->lvalue == 0)
-		do_strip = 0;
-	if ((tmp = pairfind(i->check, PW_STRIP_USER_NAME)) != NULL
-	     && tmp->lvalue == 0)
-		do_strip = 0;
-
-	if (do_strip) {
-		tmp = pairfind(request_pairs, PW_STRIPPED_USER_NAME);
-		if (tmp) {
-			strcpy((char *)tmp->strvalue, newname);
-			tmp->length = strlen((char *)tmp->strvalue);
-		} else {
-			/*
-			 *	No Stripped-User-Name exists: add one.
-			 */
-			tmp = paircreate(PW_STRIPPED_USER_NAME, PW_TYPE_STRING);
-			if (!tmp) {
-				radlog(L_ERR|L_CONS, "no memory");
-				exit(1);
-			}
-			strcpy((char *)tmp->strvalue, newname);
-			tmp->length = strlen((char *)tmp->strvalue);
-			pairadd(&request_pairs, tmp);
-		}
-		request->username = tmp;
-	}
-
 	/*
 	 *	Now add all attributes to the request list,
-	 *	except the PW_STRIP_USER_NAME one.
+	 *	except the PW_STRIP_USER_NAME one, and
+	 *	xlat them.
 	 */
 	pairdelete(&add, PW_STRIP_USER_NAME);
-	for(last = request_pairs; last && last->next; last = last->next)
-		;
-	if (last) last->next = add;
+	pairxlatmove(request, &request->packet->vps, &add);
+	pairfree(&add);
 
 	return RLM_MODULE_UPDATED;
 }
 
-/*
- *	See if the huntgroup matches. This function is
- *	tied to the "Huntgroup" keyword.
- */
-static int huntgroup_cmp(void *instance, REQUEST *req, VALUE_PAIR *request, VALUE_PAIR *check,
-			 VALUE_PAIR *check_pairs, VALUE_PAIR **reply_pairs)
-{
-	PAIR_LIST	*i;
-	char		*huntgroup;
-	rlm_preprocess_t *data = (rlm_preprocess_t *) instance;
-
-	check_pairs = check_pairs; /* shut the compiler up */
-	reply_pairs = reply_pairs;
-
-	huntgroup = (char *)check->strvalue;
-
-	for (i = data->huntgroups; i; i = i->next) {
-		if (strcmp(i->name, huntgroup) != 0)
-			continue;
-		if (paircmp(req, request, i->check, NULL) == 0) {
-			DEBUG2("  huntgroups: Matched %s at %d",
-			       i->name, i->lineno);
-			break;
-		}
-	}
-
-	/*
-	 *	paircmp() expects to see zero on match, so let's
-	 *	keep it happy.
-	 */
-	if (i == NULL) {
-		return -1;
-	}
-	return 0;
-}
-
-
 /*
  *	See if we have access to the huntgroup.
  */
-static int huntgroup_access(PAIR_LIST *huntgroups, VALUE_PAIR *request_pairs)
+static int huntgroup_access(REQUEST *request,
+			    PAIR_LIST *huntgroups, VALUE_PAIR *request_pairs)
 {
 	PAIR_LIST	*i;
 	int		r = RLM_MODULE_OK;
@@ -524,14 +348,14 @@ static int huntgroup_access(PAIR_LIST *huntgroups, VALUE_PAIR *request_pairs)
 		/*
 		 *	See if this entry matches.
 		 */
-		if (paircmp(NULL, request_pairs, i->check, NULL) != 0)
+		if (paircmp(request, request_pairs, i->check, NULL) != 0)
 			continue;
 
 		/*
 		 *	Now check for access.
 		 */
 		r = RLM_MODULE_REJECT;
-		if (hunt_paircmp(request_pairs, i->reply) == 0) {
+		if (hunt_paircmp(request, request_pairs, i->reply) == 0) {
 			VALUE_PAIR *vp;
 
 			/*
@@ -544,9 +368,9 @@ static int huntgroup_access(PAIR_LIST *huntgroups, VALUE_PAIR *request_pairs)
 						PW_TYPE_STRING);
 				if (!vp) {
 					radlog(L_ERR, "No memory");
-					exit(1);
+					r = RLM_MODULE_FAIL;
 				}
-				
+
 				strNcpy(vp->strvalue, i->name,
 					sizeof(vp->strvalue));
 				vp->length = strlen(vp->strvalue);
@@ -565,7 +389,7 @@ static int huntgroup_access(PAIR_LIST *huntgroups, VALUE_PAIR *request_pairs)
  *	If the NAS wasn't smart enought to add a NAS-IP-Address
  *	to the request, then add it ourselves.
  */
-static void add_nas_attr(REQUEST *request)
+static int add_nas_attr(REQUEST *request)
 {
 	VALUE_PAIR *nas;
 
@@ -574,7 +398,7 @@ static void add_nas_attr(REQUEST *request)
 		nas = paircreate(PW_NAS_IP_ADDRESS, PW_TYPE_IPADDR);
 		if (!nas) {
 			radlog(L_ERR, "No memory");
-			exit(1);
+			return -1;
 		}
 		nas->lvalue = request->packet->src_ipaddr;
 		ip_hostname(nas->strvalue, sizeof(nas->strvalue), nas->lvalue);
@@ -595,11 +419,12 @@ static void add_nas_attr(REQUEST *request)
 	nas = paircreate(PW_CLIENT_IP_ADDRESS, PW_TYPE_IPADDR);
 	if (!nas) {
 	  radlog(L_ERR, "No memory");
-	  exit(1);
+	  return -1;
 	}
 	nas->lvalue = request->packet->src_ipaddr;
 	ip_hostname(nas->strvalue, sizeof(nas->strvalue), nas->lvalue);
 	pairadd(&request->packet->vps, nas);
+	return 0;
 }
 
 
@@ -648,11 +473,6 @@ static int preprocess_instantiate(CONF_SECTION *conf, void **instance)
 		return -1;
 	}
 
-	/*
-	 *	Register the huntgroup comparison operation.
-	 */
-	paircompare_register(PW_HUNTGROUP_NAME, 0, huntgroup_cmp, data);
-
 	/*
 	 *	Save the instantiation data for later.
 	 */
@@ -667,6 +487,7 @@ static int preprocess_instantiate(CONF_SECTION *conf, void **instance)
 static int preprocess_authorize(void *instance, REQUEST *request)
 {
 	char buf[1024];
+	int r;
 	rlm_preprocess_t *data = (rlm_preprocess_t *) instance;
 
 	/*
@@ -700,7 +521,9 @@ static int preprocess_authorize(void *instance, REQUEST *request)
 	 *	the Request-Src-IP-Address to be used for huntgroup
 	 *	comparisons.
 	 */
-	add_nas_attr(request);
+	if (add_nas_attr(request) < 0) {
+		return RLM_MODULE_FAIL;
+	}
 
 	hints_setup(data->hints, request);
 
@@ -715,18 +538,19 @@ static int preprocess_authorize(void *instance, REQUEST *request)
 		vp = paircreate(PW_CHAP_CHALLENGE, PW_TYPE_OCTETS);
 		if (!vp) {
 			radlog(L_ERR|L_CONS, "no memory");
-			exit(1);
+			return RLM_MODULE_FAIL;
 		}
 		vp->length = AUTH_VECTOR_LEN;
 		memcpy(vp->strvalue, request->packet->vector, AUTH_VECTOR_LEN);
 		pairadd(&request->packet->vps, vp);
 	}
 
-	if (huntgroup_access(data->huntgroups, request->packet->vps) != RLM_MODULE_OK) {
+	if ((r = huntgroup_access(request, data->huntgroups,
+			     request->packet->vps)) != RLM_MODULE_OK) {
 		radlog(L_AUTH, "No huntgroup access: [%s] (%s)",
 		    request->username->strvalue,
 		    auth_name(buf, sizeof(buf), request, 1));
-		return RLM_MODULE_REJECT;
+		return r;
 	}
 
 	return RLM_MODULE_OK; /* Meaning: try next authorization module */
@@ -757,7 +581,9 @@ static int preprocess_preaccounting(void *instance, REQUEST *request)
 	/*
 	 *  Ensure that we log the NAS IP Address in the packet.
 	 */
-	add_nas_attr(request);
+	if (add_nas_attr(request) < 0) {
+		return RLM_MODULE_FAIL;
+	}
 
 	r = hints_setup(data->hints, request);
 
@@ -771,7 +597,6 @@ static int preprocess_detach(void *instance)
 {
 	rlm_preprocess_t *data = (rlm_preprocess_t *) instance;
 
-	paircompare_unregister(PW_HUNTGROUP_NAME, huntgroup_cmp);
 	pairlist_free(&(data->huntgroups));
 	pairlist_free(&(data->hints));
 
