@@ -1,5 +1,5 @@
 /*
- * $Id: db_postgresql.sql,v 1.9.2.12 2003/08/28 11:39:32 pnixon Exp $
+ * $Id: db_postgresql.sql,v 1.16 2004/02/29 13:06:58 kkalev Exp $
  *
  * Postgresql schema for FreeRADIUS
  *
@@ -167,3 +167,53 @@ CREATE TABLE nas (
 	naslocation	VARCHAR(32)
 );
 
+--
+-- Table structure for table 'radpostauth'
+--
+
+CREATE TABLE radpostauth (
+	id		BIGSERIAL PRIMARY KEY,
+	username	VARCHAR(64) NOT NULL,
+	pass		VARCHAR(64),
+	reply		VARCHAR(32),
+	authdate	TIMESTAMP with time zone NOT NULL default 'now'
+) ;
+
+--
+-- Table structure for table 'dictionary'
+-- This is not currently used by FreeRADIUS
+--
+-- CREATE TABLE dictionary (
+--     id              SERIAL PRIMARY KEY,
+--     Type            VARCHAR(30),
+--     Attribute       VARCHAR(64),
+--     Value           VARCHAR(64),
+--     Format          VARCHAR(20),
+--     Vendor          VARCHAR(32)
+-- );
+
+/*
+ * Note: (pnixon: 2003-12-10) The following function should not be required
+ * if you use the PG specific queries in raddb/postgresql.conf
+ *
+ * Common utility function for date calculations. This is used in our
+ * alternative account stop query to calculate the start of a session.
+ *
+ * This function is Copyright 2001 by Mark Steele (msteele@inet-interactif.com)
+ *
+ * Please note that this requires the plpgsql to be available in your
+ * radius database. If it is not available you can register it with
+ * postgres by running this command:
+ *
+ *   createlang plpgsql <databasename>
+ */
+CREATE FUNCTION DATE_SUB(date,int4,text) RETURNS DATE AS '
+DECLARE
+        var1 date;
+        var2 text;
+BEGIN
+        var2 = $2 || '' '' || $3;
+        SELECT INTO var1
+                to_date($1 - var2::interval, ''YYYY-MM-DD'');
+RETURN var1;
+END;' LANGUAGE 'plpgsql';
