@@ -2,7 +2,7 @@
  * rlm_sql.c		SQL Module
  * 		Main SQL module file. Most ICRADIUS code is located in sql.c
  *
- * Version:	$Id: rlm_sql.c,v 1.116.2.3 2003/09/28 13:21:09 phampson Exp $
+ * Version:	$Id: rlm_sql.c,v 1.131 2004/05/25 06:56:43 phampson Exp $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@
  */
 
 static const char rcsid[] =
-	"$Id: rlm_sql.c,v 1.116.2.3 2003/09/28 13:21:09 phampson Exp $";
+	"$Id: rlm_sql.c,v 1.131 2004/05/25 06:56:43 phampson Exp $";
 
 #include "autoconf.h"
 
@@ -48,6 +48,8 @@ static const char rcsid[] =
 #include "rlm_sql.h"
 #include "rad_assert.h"
 
+static char *allowed_chars = NULL;
+
 static CONF_PARSER module_config[] = {
 	{"driver",PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,sql_driver), NULL, "mysql"},
@@ -83,6 +85,8 @@ static CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,sqltrace), NULL, "no"},
 	{"sqltracefile", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,tracefile), NULL, SQLTRACEFILE},
+	{"readclients", PW_TYPE_BOOLEAN,
+	 offsetof(SQL_CONFIG,do_clients), NULL, "no"},
 	{"deletestalesessions", PW_TYPE_BOOLEAN,
 	 offsetof(SQL_CONFIG,deletestalesessions), NULL, "no"},
 	{"num_sql_socks", PW_TYPE_INTEGER,
@@ -105,6 +109,8 @@ static CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,accounting_onoff_query), NULL, ""},
 	{"accounting_update_query", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,accounting_update_query), NULL, ""},
+	{"accounting_update_query_alt", PW_TYPE_STRING_PTR,
+	 offsetof(SQL_CONFIG,accounting_update_query_alt), NULL, ""},
 	{"accounting_start_query", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,accounting_start_query), NULL, ""},
 	{"accounting_start_query_alt", PW_TYPE_STRING_PTR,
@@ -121,6 +127,13 @@ static CONF_PARSER module_config[] = {
 	 offsetof(SQL_CONFIG,simul_count_query), NULL, ""},
 	{"simul_verify_query", PW_TYPE_STRING_PTR,
 	 offsetof(SQL_CONFIG,simul_verify_query), NULL, ""},
+	{"postauth_table", PW_TYPE_STRING_PTR,
+	 offsetof(SQL_CONFIG,sql_postauth_table), NULL, "radpostauth"},
+	{"postauth_query", PW_TYPE_STRING_PTR,
+	 offsetof(SQL_CONFIG,postauth_query), NULL, ""},
+	{"safe-characters", PW_TYPE_STRING_PTR,
+	 offsetof(SQL_CONFIG,allowed_chars), NULL, 
+	"@abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-_: /"},
 
 	{NULL, -1, 0, NULL, NULL}
 };
@@ -144,6 +157,7 @@ static int rlm_sql_init(void) {
  *	Yucky prototype.
  */
 static int sql_set_user(SQL_INST *inst, REQUEST *request, char *sqlusername, const char *username);
+static int generate_sql_clients(SQL_INST *inst);
 
 /*
  *	sql xlat function. Right now only SELECTs are supported. Only
@@ -233,13 +247,166 @@ static int sql_xlat(void *instance, REQUEST *request,
 	return ret;
 }
 
+static int generate_sql_clients(SQL_INST *inst)
+{
+	SQLSOCK *sqlsocket;
+	SQL_ROW row;
+	char querystr[MAX_QUERY_LEN];
+	RADCLIENT *c;
+	char *netmask;
+	unsigned int i = 0;
+	
+	DEBUG("rlm_sql (%s): - generate_sql_clients",inst->config->xlat_name);
+
+	if (inst->config->sql_nas_table == NULL){
+		radlog(L_ERR, "rlm_sql (%s): sql_nas_table is NULL.",inst->config->xlat_name);
+		return -1;
+	}
+	snprintf(querystr,MAX_QUERY_LEN - 1,"SELECT * FROM %s",inst->config->sql_nas_table);
+
+	DEBUG("rlm_sql (%s): Query: %s",inst->config->xlat_name,querystr);
+	sqlsocket = sql_get_socket(inst);
+	if (sqlsocket == NULL)
+		return -1;
+	if (rlm_sql_select_query(sqlsocket,inst,querystr)){
+		radlog(L_ERR, "rlm_sql (%s): database query error, %s: %s",
+			inst->config->xlat_name,querystr,
+			(char *)(inst->module->sql_error)(sqlsocket, inst->config));
+		sql_release_socket(inst,sqlsocket);
+		return -1;
+	}
+
+	while(rlm_sql_fetch_row(sqlsocket, inst) == 0) {
+		i++;
+		row = sqlsocket->row;
+		if (row == NULL)
+			break;
+/*
+ * Format:
+ * Row1	Row2	Row3		Row4	Row5	Row6	Row7		Row8
+ *
+ * id	nasname	shortname	type	ports	secret	community	description
+ *
+ */
+
+		if (!row[0]){
+			radlog(L_ERR, "rlm_sql (%s): No row id found on pass %d",inst->config->xlat_name,i);
+			continue;
+		}
+		if (!row[1]){
+			radlog(L_ERR, "rlm_sql (%s): No nasname found for row %s",inst->config->xlat_name,row[0]);
+			continue;
+		}
+		if (strlen(row[1]) >= sizeof(c->longname)){
+			radlog(L_ERR, "rlm_sql (%s): nasname of length %d is greater than the allowed maximum of %d",
+				inst->config->xlat_name,strlen(row[1]),sizeof(c->longname) - 1);
+			continue;
+		}	
+		
+		if (!row[2]){
+			radlog(L_ERR, "rlm_sql (%s): No short name found for row %s",inst->config->xlat_name,row[0]);
+			continue;
+		}
+		if (strlen(row[2]) >= sizeof(c->shortname)){
+			radlog(L_ERR, "rlm_sql (%s): shortname of length %d is greater than the allowed maximum of %d",
+				inst->config->xlat_name,strlen(row[2]),sizeof(c->shortname) - 1);
+			continue;
+		}
+		if (row[3] && strlen(row[3]) >= sizeof(c->nastype)){
+			radlog(L_ERR, "rlm_sql (%s): nastype of length %d is greater than the allowed maximum of %d",
+				inst->config->xlat_name,strlen(row[3]),sizeof(c->nastype) - 1);
+			continue;
+		}
+		if (!row[5]){
+			radlog(L_ERR, "rlm_sql (%s): No secret found for row %s",inst->config->xlat_name,row[0]);
+			continue;
+		}
+		if (strlen(row[5]) >= sizeof(c->secret)){
+			radlog(L_ERR, "rlm_sql (%s): secret of length %d is greater than the allowed maximum of %d",
+				inst->config->xlat_name,strlen(row[5]),sizeof(c->secret) - 1);
+			continue;
+		}
+
+		DEBUG("rlm_sql (%s): Read entry nasname=%s,shortname=%s,secret=%s",inst->config->xlat_name,
+			row[1],row[2],row[5]);
+
+		c = rad_malloc(sizeof(RADCLIENT));
+		memset(c, 0, sizeof(RADCLIENT));
+
+		c->netmask = ~0;
+		netmask = strchr(row[1], '/');
+		
+		/*
+		 *      Look for netmasks.
+		 */
+		c->netmask = ~0;
+		if (netmask) {
+			int mask_length;
+
+			mask_length = atoi(netmask + 1);
+			if ((mask_length < 0) || (mask_length > 32)) {
+				radlog(L_ERR, "rlm_sql (%s): Invalid value '%s' for IP network mask for nasname %s.",
+						inst->config->xlat_name, netmask + 1,row[1]);
+				free(c);
+				continue;
+			}
+
+			if (mask_length == 0) {
+				c->netmask = 0;
+			} else {
+				c->netmask = ~0 << (32 - mask_length);
+			}
+
+			*netmask = '\0';
+			c->netmask = htonl(c->netmask);
+		}
+
+		c->ipaddr = ip_getaddr(row[1]);
+		if (c->ipaddr == INADDR_NONE) {
+			radlog(L_CONS|L_ERR, "rlm_sql (%s): Failed to look up hostname %s",
+					inst->config->xlat_name, row[1]);
+			free(c);
+			continue;
+		}
+
+		/*
+		 *      Update the client name again...
+		 */
+		if (netmask) {
+			*netmask = '/';
+			c->ipaddr &= c->netmask;
+			strcpy(c->longname, row[1]);
+		} else {
+			ip_hostname(c->longname, sizeof(c->longname),
+					c->ipaddr);
+		}
+
+		strcpy((char *)c->secret, row[5]);
+		strcpy(c->shortname, row[2]);
+		if(row[3] != NULL)
+			strcpy(c->nastype, row[3]);
+
+		DEBUG("rlm_sql (%s): Adding client %s (%s) to clients list",inst->config->xlat_name,
+			c->longname,c->shortname);
+
+		c->next = mainconfig.clients;
+		mainconfig.clients = c;
+
+	}
+	(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
+	sql_release_socket(inst, sqlsocket);
+
+	return 0;
+}
+
+
 /*
  *	Translate the SQL queries.
  */
 static int sql_escape_func(char *out, int outlen, const char *in)
 {
 	int len = 0;
-	
+
 	while (in[0]) {
 		/*
 		 *  Only one byte left.
@@ -247,13 +414,13 @@ static int sql_escape_func(char *out, int outlen, const char *in)
 		if (outlen <= 1) {
 			break;
 		}
-		
+
 		/*
 		 *	Non-printable characters get replaced with their
 		 *	mime-encoded equivalents.
 		 */
 		if ((in[0] < 32) ||
-		    strchr("@abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-_: =/", *in) == NULL) {
+		    strchr(allowed_chars, *in) == NULL) {
 			snprintf(out, outlen, "=%02X", (unsigned char) in[0]);
 			in++;
 			out += 3;
@@ -261,7 +428,7 @@ static int sql_escape_func(char *out, int outlen, const char *in)
 			len += 3;
 			continue;
 		}
-		
+
 		/*
 		 *	Else it's a nice character.
 		 */
@@ -426,7 +593,7 @@ static int rlm_sql_detach(void *instance)
 			if (module_config[i].type != PW_TYPE_STRING_PTR) {
 				continue;
 			}
-			
+
 			/*
 			 *	Treat 'config' as an opaque array of bytes,
 			 *	and take the offset into it.  There's a
@@ -535,6 +702,15 @@ static int rlm_sql_instantiate(CONF_SECTION * conf, void **instance)
 	}
 	paircompare_register(PW_SQL_GROUP, PW_USER_NAME, sql_groupcmp, inst);
 
+	if (inst->config->do_clients){
+		if (generate_sql_clients(inst) == -1){
+			radlog(L_ERR, "rlm_sql (%s): generate_sql_clients() returned error",inst->config->xlat_name);
+			rlm_sql_detach(inst);
+			return -1;
+		}
+	}
+	allowed_chars = inst->config->allowed_chars;
+
 	*instance = inst;
 
 	return RLM_MODULE_OK;
@@ -628,6 +804,8 @@ static int rlm_sql_authorize(void *instance, REQUEST * request)
 		radlog(L_DBG, "rlm_sql (%s): User %s not found in radgroupcheck",
 		       inst->config->xlat_name, sqlusername);
 	if (found || (!found && inst->config->query_on_not_found)){
+		int def_found = 0;
+
 		/*
 	 	* Check for a default_profile or for a User-Profile.
 		*/
@@ -645,7 +823,9 @@ static int rlm_sql_authorize(void *instance, REQUEST * request)
 				}
 				radius_xlat(querystr, sizeof(querystr), inst->config->authorize_group_check_query,
 									request, sql_escape_func);
-				found = sql_getvpdata(inst, sqlsocket, &check_tmp, querystr, PW_VP_GROUPDATA);
+				def_found = sql_getvpdata(inst, sqlsocket, &check_tmp, querystr, PW_VP_GROUPDATA);
+				if (def_found)
+					found = 1;
 				radius_xlat(querystr, sizeof(querystr), inst->config->authorize_group_reply_query,
 									request, sql_escape_func);
 				sql_getvpdata(inst, sqlsocket, &reply_tmp, querystr, PW_VP_GROUPDATA);
@@ -670,7 +850,7 @@ static int rlm_sql_authorize(void *instance, REQUEST * request)
 	DEBUG2("rlm_sql:  check items");
 	vp_listdebug(check_tmp);
 	DEBUG2("rlm_sql:  reply items");
-	vp_listdebug(reply_tmp); 
+	vp_listdebug(reply_tmp);
 	*/
 
 	if (paircmp(request, request->packet->vps, check_tmp, &reply_tmp) != 0) {
@@ -684,8 +864,8 @@ static int rlm_sql_authorize(void *instance, REQUEST * request)
 		return RLM_MODULE_NOTFOUND;
 	}
 
-	pairmove(&request->reply->vps, &reply_tmp);
-	pairmove(&request->config_items, &check_tmp);
+	pairxlatmove(request, &request->reply->vps, &reply_tmp);
+	pairxlatmove(request, &request->config_items, &check_tmp);
 	pairfree(&reply_tmp);
 	pairfree(&check_tmp);
 
@@ -731,7 +911,7 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 	switch (acctstatustype) {
 			/*
 			 * The Terminal server informed us that it was rebooted
-			 * STOP all records from this NAS 
+			 * STOP all records from this NAS
 			 */
 		case PW_STATUS_ACCOUNTING_ON:
 		case PW_STATUS_ACCOUNTING_OFF:
@@ -772,14 +952,36 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 				return(RLM_MODULE_FAIL);
 			if (*querystr) { /* non-empty query */
 				if (rlm_sql_query(sqlsocket, inst, querystr)) {
-					radlog(L_ERR, "rlm_sql (%s): Couldn't update SQL accounting for ALIVE packet - %s",
+					radlog(L_ERR, "rlm_sql (%s): Couldn't update SQL accounting ALIVE record - %s",
 					       inst->config->xlat_name,
 					       (char *)(inst->module->sql_error)(sqlsocket, inst->config));
 					ret = RLM_MODULE_FAIL;
 				}
+				else {
+					numaffected = (inst->module->sql_affected_rows)(sqlsocket, inst->config);
+					if (numaffected < 1) {
+
+						/*
+						 * If our update above didn't match anything
+						 * we assume it's because we haven't seen a
+						 * matching Start record.  So we have to
+						 * insert this update rather than do an update
+						 */
+						radius_xlat(querystr, sizeof(querystr), inst->config->accounting_update_query_alt, request, sql_escape_func);
+						query_log(request, inst, querystr);
+						if (*querystr) { /* non-empty query */
+							if (rlm_sql_query(sqlsocket, inst, querystr)) {
+								radlog(L_ERR, "rlm_sql (%s): Couldn't insert SQL accounting ALIVE record - %s",
+									   inst->config->xlat_name,
+									   (char *)(inst->module->sql_error)(sqlsocket, inst->config));
+								ret = RLM_MODULE_FAIL;
+							}
+							(inst->module->sql_finish_query)(sqlsocket, inst->config);
+						}
+					}
+				}
 				(inst->module->sql_finish_query)(sqlsocket, inst->config);
 			}
-
 			break;
 
 			/*
@@ -800,13 +1002,13 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 				return(RLM_MODULE_FAIL);
 			if (*querystr) { /* non-empty query */
 				if (rlm_sql_query(sqlsocket, inst, querystr)) {
-					radlog(L_ERR, "rlm_sql (%s): Couldn't update SQL accounting" " for START packet - %s",
+					radlog(L_ERR, "rlm_sql (%s): Couldn't insert SQL accounting START record - %s",
 					       inst->config->xlat_name,
 					       (char *)(inst->module->sql_error)(sqlsocket, inst->config));
 
 					/*
-					 * We failed the insert above.  It's probably because 
-					 * the stop record came before the start.  We try an
+					 * We failed the insert above.  It's probably because
+					 * the stop record came before the start.  We try
 					 * our alternate query now (typically an UPDATE)
 					 */
 					radius_xlat(querystr, sizeof(querystr), inst->config->accounting_start_query_alt, request, sql_escape_func);
@@ -814,8 +1016,8 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 
 					if (*querystr) { /* non-empty query */
 						if (rlm_sql_query(sqlsocket, inst, querystr)) {
-							radlog(L_ERR, "rlm_sql (%s): Couldn't update SQL" "accounting START record - %s",
-							       inst->config->xlat_name, 
+							radlog(L_ERR, "rlm_sql (%s): Couldn't update SQL accounting START record - %s",
+							       inst->config->xlat_name,
 							       (char *)(inst->module->sql_error)(sqlsocket, inst->config));
 							ret = RLM_MODULE_FAIL;
 						}
@@ -854,7 +1056,7 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 					if (numaffected < 1) {
 						/*
 						 * If our update above didn't match anything
-						 * we assume it's because we haven't seen a 
+						 * we assume it's because we haven't seen a
 						 * matching Start record.  So we have to
 						 * insert this stop rather than do an update
 						 */
@@ -867,7 +1069,7 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 					         */
 					        if ((pair = pairfind(request->packet->vps, PW_ACCT_SESSION_TIME)) != NULL)
 					                acctsessiontime = pair->lvalue;
-	
+
 					        if (acctsessiontime <= 0) {
 				        	        radius_xlat(logstr, sizeof(logstr), "rlm_sql: Stop packet with zero session length.  (user '%{User-Name}', nas '%{NAS-IP-Address}')", request, sql_escape_func);
 					                radlog(L_ERR, logstr);
@@ -881,7 +1083,9 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 
 						if (*querystr) { /* non-empty query */
 							if (rlm_sql_query(sqlsocket, inst, querystr)) {
-								radlog(L_ERR, "rlm_sql: Couldn't insert SQL accounting STOP record - %s", (char *)(inst->module->sql_error)(sqlsocket, inst->config));
+								radlog(L_ERR, "rlm_sql (%s): Couldn't insert SQL accounting STOP record - %s",
+										inst->config->xlat_name,
+										(char *)(inst->module->sql_error)(sqlsocket, inst->config));
 								ret = RLM_MODULE_FAIL;
 							}
 							(inst->module->sql_finish_query)(sqlsocket, inst->config);
@@ -911,7 +1115,7 @@ static int rlm_sql_accounting(void *instance, REQUEST * request) {
 /*
  *        See if a user is already logged in. Sets request->simul_count to the
  *        current session count for this user.
- * 
+ *
  *        Check twice. If on the first pass the user exceeds his
  *        max. number of logins, do a second pass and validate all
  *        logins by querying the terminal server (using eg. SNMP).
@@ -1004,7 +1208,7 @@ static int rlm_sql_checksimul(void *instance, REQUEST * request) {
         if ((vp = pairfind(request->packet->vps, PW_FRAMED_IP_ADDRESS)) != NULL)
                 ipno = vp->lvalue;
         if ((vp = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL)
-                call_num = vp->strvalue;        
+                call_num = vp->strvalue;
 
 
 	while (rlm_sql_fetch_row(sqlsocket, inst) == 0) {
@@ -1067,9 +1271,9 @@ static int rlm_sql_checksimul(void *instance, REQUEST * request) {
 				if (strcmp(row[7],"SLIP") == 0)
 					proto = 'S';
 
-			session_zap(request->packet->sockfd,
-			nas_addr,nas_port,row[2],row[1],
-			framed_addr, proto,0);
+			session_zap(request,
+				    nas_addr,nas_port,row[2],row[1],
+				    framed_addr, proto);
 		}
 	}
 
@@ -1082,6 +1286,51 @@ static int rlm_sql_checksimul(void *instance, REQUEST * request) {
 
 }
 
+/*
+ *	Execute postauth_query after authentication
+ */
+static int rlm_sql_postauth(void *instance, REQUEST *request) {
+	SQLSOCK 	*sqlsocket = NULL;
+	SQL_INST	*inst = instance;
+	char		querystr[MAX_QUERY_LEN];
+	char		sqlusername[2*MAX_STRING_LEN+10];
+
+	DEBUG("rlm_sql (%s): Processing sql_postauth", inst->config->xlat_name);
+
+	if(sql_set_user(inst, request, sqlusername, 0) <0)
+		return RLM_MODULE_FAIL;
+
+	/* If postauth_query is not defined, we stop here */
+	if (inst->config->postauth_query[0] == '\0')
+		return RLM_MODULE_NOOP;
+
+	/* Expand variables in the query */
+	memset(querystr, 0, MAX_QUERY_LEN);
+	radius_xlat(querystr, sizeof(querystr), inst->config->postauth_query,
+		    request, sql_escape_func);
+	query_log(request, inst, querystr);
+	DEBUG2("rlm_sql (%s) in sql_postauth: query is %s",
+	       inst->config->xlat_name, querystr);
+
+	/* Initialize the sql socket */
+	sqlsocket = sql_get_socket(inst);
+	if (sqlsocket == NULL)
+		return RLM_MODULE_FAIL;
+
+	/* Process the query */
+	if (rlm_sql_query(sqlsocket, inst, querystr)) {
+		radlog(L_ERR, "rlm_sql (%s) in sql_postauth: Database query error - %s",
+		       inst->config->xlat_name,
+		       (char *)(inst->module->sql_error)(sqlsocket, inst->config));
+		sql_release_socket(inst, sqlsocket);
+		return RLM_MODULE_FAIL;
+	}
+	(inst->module->sql_finish_query)(sqlsocket, inst->config);
+
+	sql_release_socket(inst, sqlsocket);
+	return RLM_MODULE_OK;
+}
+
 /* globally exported name */
 module_t rlm_sql = {
 	"SQL",
@@ -1096,7 +1345,7 @@ module_t rlm_sql = {
 		rlm_sql_checksimul,	/* checksimul */
 		NULL,			/* pre-proxy */
 		NULL,			/* post-proxy */
-		NULL			/* post-auth */
+		rlm_sql_postauth	/* post-auth */
 	},
 	rlm_sql_detach,		/* detach */
 	rlm_sql_destroy,	/* destroy */
